# --------------------------------------------------------------------------- #
#   Stylgebra                                                                 #
#                                                                             #
#   Copyright (c) 2020-2023 Steve Kieffer                                     #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at                                   #
#                                                                             #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #

from sympy.core import (
    S, Symbol, Add, Mul, Pow,
    Integer, Rational,
)

from stylgebra.excep import FormalExcep
from stylgebra.expressions import *


class SymPyExprVisitor:

    def __init__(self):
        pass

    def visit(self, expr):
        expr_name = expr.func.__name__
        method_name = f'visit_{expr_name}'
        method = getattr(self, method_name)
        if method is None:
            msg = f'Sorry, no visitor method yet for Expr of type {expr_name}'
            raise FormalExcep(msg)
        return method(expr)

    def visit_NegativeOne(self, expr):
        return self.visit_Integer(expr)

    def visit_Symbol(self, expr):
        assert isinstance(expr, Symbol)
        return FormalVariable(expr.name)

    def visit_Integer(self, expr):
        assert isinstance(expr, Integer)
        return FormalInteger(value=int(expr))

    def visit_Rational(self, expr):
        assert isinstance(expr, Rational)
        return formalQuoFromFrac(expr)

    def visit_Add(self, expr):
        assert isinstance(expr, Add)
        return FormalSum(terms=[
            self.visit(term) for term in expr.args
        ])

    def visit_Mul(self, expr):
        assert isinstance(expr, Mul)
        args = expr.args

        if len(args) == 2 and args[0] == S.NegativeOne:
            term = self.visit(args[1])
            return FormalSummand(term, sign=-1)

        upstairs = []
        downstairs = []
        for arg in args:
            if isinstance(arg, Pow) and arg.args[1] < 0:
                inverse = Pow(arg.base, -arg.exp)
                downstairs.append(self.visit(inverse))
            else:
                upstairs.append(self.visit(arg))

        def handle_factor_list(factors):
            if len(factors) > 1:
                return FormalProduct(factors=factors)
            elif factors:
                return factors[0]
            else:
                return self.visit(S(1))

        u = handle_factor_list(upstairs)
        if downstairs:
            d = handle_factor_list(downstairs)
            return FormalQuotient(u, d)
        else:
            return u

    def visit_Pow(self, expr):
        assert isinstance(expr, Pow)
        b = self.visit(expr.base)
        e = self.visit(expr.exp)
        return FormalPower(b, e)


def adapt_sympy_expr(expr):
    """
    Turn a SymPy Expr into one of our formal class instances.

    :param expr: a sympy Expr instance
    :return: a Formal instance
    """
    visitor = SymPyExprVisitor()
    formal = visitor.visit(expr)
    return formal



# --------------------------------------------------------------------------- #
#   Stylgebra                                                                 #
#                                                                             #
#   Copyright (c) 2020-2023 Steve Kieffer                                     #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at                                   #
#                                                                             #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #

from sympy import S, Add, symbols

from stylgebra.adapt import adapt_sympy_expr
from tests.conftest import check


def test_adapt_00():
    e = S.NegativeOne
    f = adapt_sympy_expr(e)
    s = f.format()
    check(s, "-1")


def test_adapt_01():
    x, y = symbols('x, y')
    a = Add(x, y, 3, evaluate=False)
    f = adapt_sympy_expr(a)
    s = f.format()
    check(s, "x + y + 3")


def test_adapt_02():
    x, y = symbols('x, y')
    a = Add(x, -y, evaluate=False)
    f = adapt_sympy_expr(a)
    s = f.format()
    check(s, "x - y")


def test_adapt_03():
    print()
    x = symbols('x')
    expr1 = x**(-1)
    expr2 = x**(-2)
    print('SymPy:')
    print(expr1)
    print(expr2)
    print('=============')

    formal1 = adapt_sympy_expr(expr1)
    formal2 = adapt_sympy_expr(expr2)

    s1a = formal1.format(rules={'Power': {'negative': 'sup'}})
    s1b = formal1.format(rules={'Power': {'negative': 'frac'}})
    s1c = formal1.format(rules={'Power': {'negative': 'inline'}})
    check(s1a, 'x^{-1}')
    check(s1b, r'\frac{1}{x}')
    check(s1c, '1/x')

    print()

    s2a = formal2.format(rules={'Power': {'negative': 'sup'}})
    s2b = formal2.format(rules={'Power': {'negative': 'frac'}})
    s2c = formal2.format(rules={'Power': {'negative': 'inline'}})
    check(s2a, 'x^{-2}')
    check(s2b, r'\frac{1}{x^{2}}')
    check(s2c, '1/x^{2}')



# --------------------------------------------------------------------------- #
#   Stylgebra                                                                 #
#                                                                             #
#   Copyright (c) 2020-2023 Steve Kieffer                                     #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at                                   #
#                                                                             #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #

import random
from fractions import Fraction


def nice_random_integer(
        digit_lh = (0.5, 0.5),
        negative_lh = 0.5,
        zero_ok = True):
    """
    Generate a "nice random" integer.

    This is for use when generating examples of structures that involve
    integers, and it is meant to make the kinds of decisions you
    would probably make while writing down examples on a chalkboard:

    You would choose both positive and negative numbers; the numbers would
    not be terribly large. This is what "nice" means.

    In the parameters, an "lh" suffix means "liklihood".

    :param digit_lh: pass a tuple of floats summing to unity. These are the
        liklihoods, respectively, of the number we generate having 1, 2, 3, ... digits.
        For example the default value of (0.5, 0.5) means that half the time we
        will generate a 1-digit number, and the other half a 2-digit number.
    :param negative_lh: liklihood of the number being negative
    :param zero_ok: if False, we will not return zero

    :return: int
    """
    # Decide how many digits the number will be.
    d = random.choices(range(1, len(digit_lh) + 1), weights=digit_lh)[0]

    # Compute the closed-open interval [a, b) from which the integer will be chosen.
    b = 10**d
    a = 0 if (d == 1 and zero_ok) else b/10

    n = random.randrange(a, b)
    if random.random() < negative_lh:
        n *= -1

    return n


def nice_random_rational(
        integer_lh = 0.5,
        digit_lh = (0.5, 0.5),
        negative_lh = 0.5,
        zero_ok = True
    ):
    """
    Generate a "nice random" rational number.

    This function simply builds off the one that generates "nice random" integers,
    and again is meant to generate pseudo-random rational numbers of the kind
    you would tend to write down at the chalkboard.

    You would choose both positive and negative numbers; several would
    be pure integers (to illustrate that that's okay); neither numerator
    nor denominator would ever be terribly large. This is what "nice" means.

    In the parameters, an "lh" suffix means "liklihood".
    Note, however, that these tend to be only approximate liklihoods in the
    end, due to a final step of reducing to least terms.

    Thus, e.g., we will start out deciding on a pure integer or proper rational
    number based on the `integer_lh` parameter. However, over time you will
    get slightly more pure integers than you asked for, since e.g. we may
    generate 10/2 and then return 5 after reducing.

    :param integer_lh: liklihood of the number being a pure integer.
    :param digit_lh: as in `nice_random_integer`
    :param negative_lh: as in `nice_random_integer`
    :param zero_ok: as in `nice_random_integer`

    :return: instance of Python's fractions.Fraction class
    """
    # Numerator:
    n = nice_random_integer(digit_lh=digit_lh, negative_lh=negative_lh, zero_ok=zero_ok)

    # Denominator:
    d = 1
    if n != 0 and random.random() >= integer_lh:
        d = nice_random_integer(digit_lh=digit_lh, negative_lh=0, zero_ok=False)

    return Fraction(n, d)



# --------------------------------------------------------------------------- #
#   Stylgebra                                                                 #
#                                                                             #
#   Copyright (c) 2020-2023 Steve Kieffer                                     #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at                                   #
#                                                                             #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #

from stylgebra.algebra import *
from tests.conftest import check


exp_algebra_00 = [
    r"\zeta",
    r"\mathrm{e}^{2\pi i/7}",
    r"a primitive $7^{\mathrm{th}}$ root of unity",
    r"\xi",
    r"a primitive root of unity",
    r"$\xi$, a primitive root of unity",
]
def test_algebra_00():
    zeta = FormalPrimitiveRootOfUnity(7)
    check(zeta.format(), exp_algebra_00[0])
    check(zeta.format(rules={
        "@": {
            'form': 'value'
        }
    }), exp_algebra_00[1])
    check(zeta.format(rules={
        "@": {
            'form': 'verbal'
        }
    }), exp_algebra_00[2])

    xi = FormalPrimitiveRootOfUnity(name=r'\xi')
    s = xi.format()
    t = xi.format(rules={
        "@": {
            'form': 'verbal'
        }
    })
    check(s, exp_algebra_00[3])
    check(t, exp_algebra_00[4])
    check(s + ', ' + t, exp_algebra_00[5])


exp_algebra_01 = [
    "K",
    r"\mathbb{Q}(\zeta)",
    r"\mathbb{Q}(\zeta_{p})",
    r"\mathbb{Q}(\zeta_{13})",
    r"\mathbb{Q}(\mathrm{e}^{2\pi i/p})",
    r"\mathbb{Q}(\mathrm{e}^{2\pi i/13})",
    r"the $p^{\mathrm{th}}$ cyclotomic field",
    r"the $13^{\mathrm{th}}$ cyclotomic field",
]
def test_algebra_01():
    p = FormalInteger(13, 'p')
    K = FormalCyclotomicField(p, name="K")

    check(K.format(), exp_algebra_01[0])

    check(K.format({
        'form': 'symbolic'
    }), exp_algebra_01[1])

    # Instead of defining a separate mode, we can put the top-level mode
    # into the rules dict, under the key "@". This matches only items whose
    # path is empty, and that will be only the top-level item.
    check(K.format(rules={
        "@": {
            'form': 'symbolic'
        },
        '@gen': {
            'form': 'symbolic'
        }
    }), exp_algebra_01[2])

    check(K.format(rules={
        "@": {
            'form': 'symbolic'
        },
        '@gen': {
            'form': 'symbolic'
        },
        "@gen-order": {
            'form': 'value'
        }
    }), exp_algebra_01[3])

    check(K.format(rules={
        "@": {
            'form': 'symbolic'
        },
        "@gen": {
            'form': 'value'
        }
    }), exp_algebra_01[4])

    check(K.format(rules={
        "@": {
            'form': 'symbolic'
        },
        "@gen": {
            'form': 'value'
        },
        "@gen-order": {
            'form': 'value'
        }
    }), exp_algebra_01[5])

    check(K.format({
        'form': 'verbal'
    }), exp_algebra_01[6])

    check(K.format(rules={
        "@": {
            'form': 'verbal'
        },
        "@order": {
            'form': 'value'
        }
    }), exp_algebra_01[7])


exp_algebra_02 = [
    r"a_{0} + a_{1} \zeta + a_{2} \zeta^{2} + \cdots + a_{p - 2} \zeta^{p - 2}",
    r"a_{0} + a_{1} \zeta + \cdots + a_{p - 2} \zeta^{p - 2}",
    r"a_{p - 2} \zeta^{p - 2} + a_{p - 3} \zeta^{p - 3} + \cdots + a_{0}",
    r"a_{0} + a_{1} \zeta + a_{2} \zeta^{2} + a_{3} \zeta^{3} + a_{4} \zeta^{4} + a_{5} \zeta^{5}",
    r"c_{0} + c_{1} \zeta + c_{2} \zeta^{2} + \cdots + c_{p - 2} \zeta^{p - 2}",
    r"c_{0} + c_{1} \zeta + c_{2} \zeta^{2} + \cdots + c_{7 - 2} \zeta^{7 - 2}",
    r"c_{0} + c_{1} \zeta + c_{2} \zeta^{2} + \cdots + c_{5} \zeta^{5}",
    r"2 + 3 \zeta + 5 \zeta^{2} + \cdots + 13 \zeta^{p - 2}",
]
def test_algebra_02():
    p = FormalInteger(7, 'p')
    K = FormalCyclotomicField(p, name="K")

    elt = K.build_formal_element()
    check(elt.format(), exp_algebra_02[0])

    elt = K.build_formal_element(elide_after=2)
    check(elt.format(), exp_algebra_02[1])

    elt = K.build_formal_element(elide_after=2, falling_powers=True)
    check(elt.format(), exp_algebra_02[2])

    elt = K.build_formal_element(elide_after=0)
    check(elt.format(), exp_algebra_02[3])

    elt = K.build_formal_element(coeff_base='c')
    check(elt.format(), exp_algebra_02[4])
    check(elt.format(rules={
        "#p": {
            'form': 'value'
        }
    }), exp_algebra_02[5])

    # The hard way:
    check(elt.format(rules={
        "@term4-factor0-sub-subst": {
            'form': 'value'
        },
        "@term4-factor1-power-subst": {
            'form': 'value'
        }
    }), exp_algebra_02[6])

    # The easy way:
    check(elt.format(rules={
        "@term4 Sum": {
            'form': 'value'
        }
    }), exp_algebra_02[6])

    # Or just address all variables; needn't confine to term4:
    check(elt.format(rules={
        "Variable @subst": {
            'form': 'value'
        }
    }), exp_algebra_02[6])

    # Or you can address the iteration variable, either by knowing that
    # by default it will be 'i', or by setting your own index:
    elt = K.build_formal_element(coeff_index=FormalVariable('j'), coeff_base='c')
    check(elt.format(rules={
        "#j @subst": {
            'form': 'value'
        }
    }), exp_algebra_02[6])

    i = FormalVariable('i')
    f = FormalLookup(i, [2, 3, 5, ..., ..., 13])
    elt = K.build_formal_element(coeff_form=f, coeff_index=i)
    check(elt.format(), exp_algebra_02[7])

    elt = K.build_formal_element(random_coeffs=True)
    print(elt.format())


exp_algebra_03 = [
    r'\left\{ a_{0} + a_{1} \zeta + a_{2} \zeta^{2} + \cdots + a_{p - 2} \zeta^{p - 2} : a_{i} \in \mathbb{Q} \right\}',
    r"\left\{ a_{0} + a_{1} \zeta + a_{2} \zeta^{2} + \cdots + a_{p - 2} \zeta^{p - 2} : \mbox{$a_{i}$ is an element of $\mathbb{Q}$} \right\}",
    r"$\left\{ a_{0} + a_{1} \zeta + a_{2} \zeta^{2} + \cdots + a_{p - 2} \zeta^{p - 2} \right\}$ where $a_{i}$ is an element of $\mathbb{Q}$",
]
def test_algebra_03():
    p = FormalInteger(7, 'p')
    K = FormalCyclotomicField(p, name="K")
    S = K.build_formal_set()

    check(S.format({
        'form': 'symbolic'
    }), exp_algebra_03[0])

    check(S.format({
        'form': 'symbolic'
    }, {
        '@cond': {
            'form': 'verbal'
        }
    }), exp_algebra_03[1])

    check(S.format({
        'form': 'symbolic',
        'cond': 'where',
    }, {
        '@cond': {
            'form': 'verbal',
        }
    }), exp_algebra_03[2])


def test_algebra_04():
    p = FormalInteger(7, 'p')
    K = FormalCyclotomicField(p, name="K")
    G = FormalGaloisGroup(K, name="G")

    check(G.format(), "G")

    check(G.format(rules={
        "@": {
            'form': 'symbolic'
        }
    }), r'\Gal(K/\mathbb{Q})')
    check(G.format(rules={
        "@": {
            'form': 'symbolic'
        },
        "@ext": {
            'form': 'symbolic'
        }
    }), r'\Gal(\mathbb{Q}(\zeta)/\mathbb{Q})')
    check(G.format(rules={
        "@": {
            'form': 'symbolic'
        },
        "@ext": {
            'form': 'symbolic'
        },
        "@ext-gen": {
            'form': 'value'
        }
    }), r'\Gal(\mathbb{Q}(\mathrm{e}^{2\pi i/p})/\mathbb{Q})')

    check(G.format(rules={
        "@": {
            'form': 'verbal'
        }
    }), r'the Galois group of $K$ over $\mathbb{Q}$')
    check(G.format(rules={
        "@": {
            'form': 'verbal'
        },
        "@ext": {
            'form': 'symbolic'
        }
    }), r'the Galois group of $\mathbb{Q}(\zeta)$ over $\mathbb{Q}$')

    check(G.format(rules={
        "@": {
            'form': 'value'
        },
        "@uset": {
            'form': 'symbolic'
        }
    }), r'\left\{ \sigma_{1}, \sigma_{2}, \ldots, \sigma_{p - 1} \right\}')

    check(G.format(rules={
        "@": {
            'form': 'value'
        },
        "@uset": {
            'form': 'symbolic',
            'range': 'bind',
        }
    }), r'\left\{ \sigma_{r} : 1 \leq r \leq p - 1 \right\}')

    check(G.format(rules={
        "@": {
            'form': 'value'
        },
        "@uset": {
            'form': 'symbolic',
            'range': 'bind',
            'cond': 'vbar',
        },
        "@uset-elt0": {
            'form': 'name-mapsto'
        }
    }), r'\left\{ \sigma_{r}: \zeta \mapsto \zeta^{r} \mid 1 \leq r \leq p - 1 \right\}')

    check(G.format(rules={
        "@": {
            'form': 'value'
        },
        "@uset": {
            'form': 'symbolic'
        },
        "@uset-elt[i]": {
            'form': 'mapsto'
        }
    }), r'\left\{ \zeta \mapsto \zeta, \zeta \mapsto \zeta^{2}, \ldots, \zeta \mapsto \zeta^{p - 1} \right\}')

    check(G.format(rules={
        "@": {
            'form': 'value'
        },
        "@uset": {
            'form': 'symbolic',
            'generator': 'auto',
            'range': 'bind',
        }
    }), r'\left\{ \sigma_{k} : 1 \leq k \leq p - 1 \right\}')

    check(G.format(rules={
        "@": {
            'form': 'value'
        },
        "@uset": {
            'form': 'symbolic',
            'generator': 'auto',
            'range': 'bind',
            'cond': 'vbar',
        },
        "@uset-elt0": {
            'form': 'name-mapsto'
        }
    }), r'\left\{ \sigma_{k}: \zeta \mapsto \zeta^{\gamma^{k}} \mid 1 \leq k \leq p - 1 \right\}')

    check(G.format(rules={
        "@": {
            'form': 'value'
        },
        "@uset": {
            'form': 'symbolic',
            'generator': 'auto',
        },
        "@uset-elt[i]": {
            'form': 'mapsto'
        }
    }), r'\left\{ \zeta \mapsto \zeta^{\gamma}, \zeta \mapsto \zeta^{\gamma^{2}}, \ldots, \zeta \mapsto \zeta^{\gamma^{p - 1}} \right\}')

    check(G.format(rules={
        "@": {
            'form': 'value'
        },
        "@uset": {
            'form': 'symbolic',
            'generator': 'auto',
        },
        "@uset-elt[i]": {
            'form': 'mapsto'
        },
        "#gamma": {
            'form': 'value'
        }
    }), r'\left\{ \zeta \mapsto \zeta^{3}, \zeta \mapsto \zeta^{3^{2}}, \ldots, \zeta \mapsto \zeta^{3^{p - 1}} \right\}')

    # Commenting out the last two tests for now.
    # The solution in keeping with our overall design would be to have a styling
    # option for instances of SymPy's `SymmetricModularInteger____` classes
    # (i.e. elements of finite fields), where you can say whether you want the
    # form `r mod m` or simply `r`.
    r"""
    check(G.format(rules={
        "@": {
            'form': 'value'
        },
        "@uset": {
            'form': 'symbolic',
            'generator': 'auto',
        },
        "@uset-elt[i]": {
            'form': 'mapsto'
        },
        "@uset-elt[i]-value-power": {
            'form': 'value'
        }
    }), r'\left\{ \zeta \mapsto \zeta^{3}, \zeta \mapsto \zeta^{2}, \ldots, \zeta \mapsto \zeta \right\}')
    
    check(G.format(rules={
        "@": {
            'form': 'value'
        },
        "@uset": {
            'form': 'symbolic',
            'generator': 5,
        },
        "@uset-elt[i]": {
            'form': 'mapsto'
        },
        "@uset-elt[i]-value-power": {
            'form': 'value'
        }
    }), r'\left\{ \zeta \mapsto \zeta^{5}, \zeta \mapsto \zeta^{4}, \ldots, \zeta \mapsto \zeta \right\}')
    """



# --------------------------------------------------------------------------- #
#   Stylgebra                                                                 #
#                                                                             #
#   Copyright (c) 2020-2023 Steve Kieffer                                     #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at                                   #
#                                                                             #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #


from sympy import primitive_root

from stylgebra.basic import *
from stylgebra.expressions import *
from stylgebra.relations import *
from stylgebra.numerical import nice_random_rational


class FormalGroup(Formal):

    def __init__(self, name=None, id=None):
        super().__init__(name=name, id=id)

    @withmode
    def format(self, mode=None, rules=None, path=None, modifier=None):
        return MM(self.name())


class FormalRing(Formal):

    def __init__(self, name=None, id=None):
        super().__init__(name=name, id=id)

    def __getitem__(self, item):
        if isinstance(item, FormalGroup):
            return FormalGroupRing(self, item)
        else:
            pass  # TODO

    @withmode
    def format(self, mode=None, rules=None, path=None, modifier=None):
        return MM(self.name())


class FormalGroupRing(FormalRing):

    def __init__(self, ring, group, name=None, id=None):
        super().__init__(name=name, id=id)
        self.ring = ring
        self.group = group


class FormalField(FormalRing):

    pass


class FormalNumberField(FormalField):

    def build_galois_group_underlying_set(self, base_field):
        raise NotImplementedError()


class TheRationalNumbers(FormalNumberField):

    def __init__(self):
        super().__init__(r'\mathbb{Q}', "QQ")

    def __call__(self, *args):
        if len(args) == 1 and isinstance(args[0], FormalPrimitiveRootOfUnity):
            zeta = args[0]
            return FormalCyclotomicField(gen=zeta)
        else:
            raise NotImplementedError()

# The Rational Numbers, as a fixed instance:
QQ = TheRationalNumbers()


class FormalGaloisGroup(FormalGroup):

    def __init__(self, ext_field, base_field=QQ, name=None, id=None):
        super().__init__(name=name, id=id)
        self.ext_field = ext_field
        self.base_field = base_field

    @withmode
    def format(self, mode=None, rules=None, path=None, modifier=None):
        """
        options:

            form: FormOpt
                Default: name
                name: self.name()
                value: ask self.ext_field to produce a representation of the
                    underlying set
                symbolic: Gal(E/F) with E = self.ext_field, F = self.base_field
                verbal: "the Galois group of .... over ..."

        subpaths:

            'ext': self.ext_field
            'base': self.base_field
            'uset': only for {'form': 'value'} case: the underlying set
        """
        form = mode.get('form', FormOpt.NAME)

        if form == FormOpt.NAME:
            return MM(self.name())
        elif form == FormOpt.VALUE:
            assert isinstance(self.ext_field, FormalNumberField)
            fwdMode = self.getFwdMode(self, 'uset', rules, path)
            uset = self.ext_field.build_galois_group_underlying_set(self.base_field, mode=fwdMode)
            return self.fwd(uset, 'uset', rules, path)

        ext = self.fwd(self.ext_field, 'ext', rules, path)
        base = self.fwd(self.base_field, 'base', rules, path)

        if form == FormOpt.SYMBOLIC:
            return MM(r'\Gal(') + ext + MM('/') + base + MM(')')
        elif form == FormOpt.VERBAL:
            return 'the Galois group of ' + ext + ' over ' + base


class FormalPrimitiveRootOfUnity(Formal):

    def __init__(self, m=None, name=r'\zeta', id=None):
        super().__init__(name=name, id=id)
        self.m = wrap(m) if m is not None else None

    def order(self):
        return self.m

    @withmode
    def format(self, mode=None, rules=None, path=None, modifier=None):
        """
        mode options:

            form: FormOpt
                name: self.name()
                value: write as a complex exponential
                symbolic: put self.m as subscript on self.name()
                verbal: describe verbally

                Default: 'name'

        subpaths:

            'order': self.m
        """
        form = mode.get('form', FormOpt.NAME)

        if form == FormOpt.NAME:
            return MM(self.name())

        if form == FormOpt.VERBAL:
            if self.m is None:
                return 'a primitive root of unity'
            else:
                m_ordinal = self.fwd(self.m, 'order', rules, path, modifier={'ordinal': True})
            return 'a primitive ' + m_ordinal + ' root of unity'

        assert self.m is not None

        order = self.fwd(self.m, 'order', rules, path)

        if form == FormOpt.SYMBOLIC:
            base = MM(self.name())
            return base + MM('_{') + order + MM('}')
        else:
            assert form == FormOpt.VALUE
            return MM(r'\mathrm{e}^{2\pi i/%s}' % order)


class FormalCyclotomicField(FormalNumberField):

    def __init__(self, m=None, gen=None, name=None, id=None):
        FormalNumberField.__init__(self, name=name, id=id)
        self.basefield = QQ
        if m is not None:
            # In this case we ignore gen.
            self.m = wrap(m)
            self.gen = FormalPrimitiveRootOfUnity(m)
        elif gen is not None:
            assert isinstance(gen, FormalPrimitiveRootOfUnity)
            self.m = wrap(gen.order())
            self.gen = gen
        else:
            # Both m and gen are None.
            self.m = m
            self.gen = gen

    def is_prime(self):
        return self.m.prime if self.m is not None else False

    @withmode
    def format(self, mode=None, rules=None, path=None, modifier=None):
        """
        mode options:

            form: FormOpt
                Default: 'name'

        subpaths:
            'gen': self.gen
            'order': self.m
            'basefield': self.basefield
        """
        form = mode.get('form', FormOpt.NAME)

        if form == FormOpt.NAME:
            return MM(self.name())

        # TODO:
        #  By analogy to what we're now doing with the GaloisGroup class,
        #  maybe 'value' should mean we return our representation of our
        #  underlying set (while 'symbolic' still means what it means here).
        #  In general, I suppose, 'value' for any algebraic structure could
        #  mean a representation of the underlying set.

        elif form in FormOpt.valsym:
            basefield = self.fwd(self.basefield, 'basefield', rules, path)
            gen = self.fwd(self.gen, 'gen', rules, path)
            return basefield + MM("(") + gen + MM(")")

        elif form == FormOpt.VERBAL:
            if self.m is None:
                return 'a cyclotomic field'
            m_ordinal = self.fwd(self.m, 'order', rules, path, modifier={'ordinal': True})
            return 'the ' + m_ordinal + ' cyclotomic field'

    def build_formal_element(
            self,
            coeff_form = None,
            coeff_index = None,
            random_coeffs = False,
            coeff_base = 'a',
            elide_after = 3,
            falling_powers = False
    ):
        """
        Build and return a formal object to represent an element of this structure.

        coeff_form, random_coeffs, coeff_base, coeff_index:
            These options have to do with setting the kind of coefficients you want.

            coeff_form: You can set whatever coefficient form you want, using coeff_form.

            coeff_index: If you pass a coeff_form then, unless it is a constant, it should
                have at least one FormalVariable occurring somewhere within it, to serve as
                the iteration index. You must pass the desired FormalVariable for coeff_index,
                so that we can substitute values for it, as we build a formal sum.

            random_coeffs: If coeff_form is None and random_coeffs is True, we will
                generate "nice random" coeffs.

            coeff_base: If coeff_form is None and random_coeffs is False, then we will
                write the coeffs as a sequence of subscripted variables, and you can choose
                the base letter in coeff_base.

        elide_after, falling_powers:
            These options control what kind of terms you get, and in what order.

            elide_after: Pass a positive integer if you want an elision (...) after that
                many initial terms. Set to 0 if you do not want an elision (i.e. you want
                to spell out all the terms).

            falling_powers: Set True if you want the terms to go from high powers of the
                generator to low, reading left to right. Otherwise they go from low to high.
        """

        if self.is_prime():

            i = FormalVariable('i') if coeff_index is None else coeff_index
            if coeff_form is None:
                if random_coeffs:
                    a = lambda k: formalQuoFromFrac(nice_random_rational())
                    coeff_form = FormalLookup(i, a)
                else:
                    a = FormalString(coeff_base)
                    coeff_form = a._(i)

            term_form = coeff_form * self.gen ** i

            mval = self.m.value()
            if mval is not None:
                # When m does have a numerical value, the number of terms before
                # elision is capped at m - 4.
                elide_after = min(elide_after, mval - 4)
            else:
                # When m does not have a numerical value, then elision is required.
                if elide_after < 1:
                    elide_after = 3

            if elide_after > 0:
                m = self.m
                if falling_powers:
                    rng = [m - (k+2) for k in range(elide_after)] + [..., 0]
                else:
                    rng = list(range(elide_after)) + [..., m - 2]
            else:
                rng = range(mval - 1)
                if falling_powers:
                    rng = reversed(rng)

            return FormalRangeSum(rng, term_form, i)

        else:
            # m is not known to be prime
            ...  # TODO
            raise NotImplementedError()


    def build_formal_set(self, **kwargs):
        """
        Build and return a formal object to represent the underlying set of
        this structure.

        kwargs: accepts the same keyword args as the `build_formal_element` method.
        """
        kwargs.update({
            'random_coeffs': False
        })
        alpha = self.build_formal_element(**kwargs)
        coeff = alpha.generic_term().factors()[0]
        cond = coeff |in_| QQ
        return FormalSet(alpha, condition=cond)

    def build_galois_group_underlying_set(self, base_field, mode=None):
        mode = mode or {}
        if self.is_prime():

            sig = FormalString(r'\sigma')

            generator = mode.get('generator')

            if generator is None:
                r = FormalVariable('r')
                elt = FormalMapping(sig._(r), self, self, [self.gen], self.gen ** r)
                uset = FormalRangeSet([1, 2, ..., self.m - 1], elt, r)

            else:
                from stylgebra.elnt import FormalIntResidue
                m = self.m.value()
                if generator == 'auto':
                    generator = primitive_root(m)
                gamma = FormalIntResidue(generator, m, name=r'\gamma', id='gamma')
                k = FormalVariable('k')
                elt = FormalMapping(sig._(k), self, self, [self.gen], self.gen ** (gamma ** k))
                uset = FormalRangeSet([1, 2, ..., self.m - 1], elt, k)

            return uset

        else:
            # m is not known to be prime
            ...  # TODO
            raise NotImplementedError()

##############################################################################

def try_00():
    p = FormalInteger(13, 'p')
    K = FormalCyclotomicField(p, name="K")
    s = K.format({
        'form': 'symbolic'
    })
    print(s)

if __name__ == "__main__":
    try_00()



# --------------------------------------------------------------------------- #
#   Stylgebra                                                                 #
#                                                                             #
#   Copyright (c) 2020-2023 Steve Kieffer                                     #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at                                   #
#                                                                             #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #

from stylgebra.expressions import *
from stylgebra.relations import *

from tests.conftest import check


def test_basic_01():
    a = FormalString('a')
    ivar = FormalVariable('i')
    ai = a._(ivar)
    check(
        [ai.format(rules={'#i': i}) for i in range(0, 5)],
        "['a_{0}', 'a_{1}', 'a_{2}', 'a_{3}', 'a_{4}']"
    )


def test_basic_02():
    p = FormalInteger(11, 'p')
    check(p.format({'form': 'name'}), 'p')
    check(p.format({'form': 'value'}), '11')
    check(p.format({'form': 'verbal'}), 'an odd, positive prime')
    check('the ' + p.format({
        'form': 'name',
        'ordinal': True
    }), r'the $p^{\mathrm{th}}$')
    check('the ' + p.format({
        'form': 'value',
        'ordinal': True
    }), r'the $11^{\mathrm{th}}$')
    check('the ' + FormalInteger(21).format({
        'form': 'value',
        'ordinal': True
    }), r'the $21^{\mathrm{st}}$')
    check('the ' + FormalInteger(1).format({
        'form': 'value',
        'ordinal': True
    }), r'the $1^{\mathrm{st}}$')
    check('the ' + FormalInteger(2).format({
        'form': 'value',
        'ordinal': True
    }), r'the $2^{\mathrm{nd}}$')
    check('the ' + FormalInteger(3).format({
        'form': 'value',
        'ordinal': True
    }), r'the $3^{\mathrm{rd}}$')


exp_basic_03 = [
    r"\zeta^{r}",
    r"\zeta^{-1}",
    r"\frac{1}{\zeta}",
]
def test_basic_03():
    zeta = FormalVariable(r'\zeta')
    r = FormalInteger(-1, 'r')
    ztor = zeta**r
    check(ztor.format({}, {'@power': {'form': 'name'}}), exp_basic_03[0])
    check(ztor.format({}, {'@power': {'form': 'value'}}), exp_basic_03[1])
    check(ztor.format({'negative': 'frac'}, {'@power': {'form': 'value'}}), exp_basic_03[2])


def test_basic_04():
    b = FormalInteger(2, 'b')
    r = FormalInteger(3, 'r')
    btor = b**r
    check(btor.format(), 'b^{r}')
    check(btor.format({}, {
        '@base': {'form': 'value'},
        '@power': {'form': 'value'}
    }), '2^{3}')
    check(btor.format({'form': 'value'}), '8')


exp_basic_05 = [
    r"\sigma_{i}",
    r"\sigma_{i}(\zeta)",
    r"\zeta^{i}",
    r"\sigma_{i}: F \rightarrow F",
    r"\sigma_{i}: \zeta \mapsto \zeta^{i}",
    r"\sigma_{5}: \zeta \mapsto \zeta^{5}",
    r"\sigma_{i}: \xi \mapsto \zeta^{i}",
    r"\sigma_{i}: \xi \mapsto \xi^{i}",
]
def test_basic_05():
    sig = FormalString(r'\sigma')
    i = FormalVariable('i')

    #p = FormalInteger(11, 'p')
    #F = CyclotomicField(p)
    F = FormalVariable("F")

    zeta = FormalVariable(r'\zeta', 'zeta')
    sigi = FormalMapping(sig._(i), F, F, [zeta], zeta ** i)

    check(sigi.format(), exp_basic_05[0])
    check(sigi.format({
        'form': 'name-args'
    }), exp_basic_05[1])
    check(sigi.format({
        'form': 'value'
    }), exp_basic_05[2])
    check(sigi.format({
        'form': 'map'
    }), exp_basic_05[3])
    check(sigi.format({
        'form': 'name-mapsto'
    }), exp_basic_05[4])
    check(sigi.format({'form': 'name-mapsto'}, {'#i': 5}), exp_basic_05[5])

    check(sigi.format({'form': 'name-mapsto'}, {
        '@arg0': FormalVariable(r'\xi').format()
    }), exp_basic_05[6])

    check(sigi.format({'form': 'name-mapsto'}, {'#zeta': r'\xi'}), exp_basic_05[7])


exp_basic_06 = [
    r"\left(a + b\right) + c",
    r"a + b + c",
    r"3",
    r"3",
    r"\left(1 + 2\right) + 0",
    r"\left(a + b\right) - \left(c + d\right)",
    r"a + b - c - d",
    r"-c + a + b - d",
    r"\left(-c + a\right) + \left(b - d\right)",
    r"-1",
    r"-1",
    r"\left(a + b\right) - d",
    r"\left(a + b\right) - \left(0 + d\right)",
    r"a + b - 0 - d",
    r"-a + b",
    r"a - b",
    r"a + b + \cdots + d",
]
def test_basic_06():
    a = FormalInteger(1, 'a')
    b = FormalInteger(2, 'b')
    c = FormalInteger(0, 'c')
    d = FormalInteger(4, 'd')
    el = FormalEllipsis()

    s0 = a + b + c
    check(s0.format(), exp_basic_06[0])
    check(s0.flattened().format(), exp_basic_06[1])
    check(s0.format({'form': 'value'}), exp_basic_06[2])
    check(s0.flattened().format({'form': 'value'}), exp_basic_06[3])
    check(s0.format({
        'show-zeros': True
    }, {
        '#a': {'form': 'value'},
        '#b': {'form': 'value'},
        '#c': {'form': 'value'},
    }), exp_basic_06[4])

    s1 = (a + b) - (c + d)
    check(s1.format(), exp_basic_06[5])
    check(s1.flattened().format(), exp_basic_06[6])
    check(s1.flattened().permuted([2, 0, 1, 3]).format(), exp_basic_06[7])
    check(s1.flattened().split([
        [2, 0], [1, 3]
    ]).format(), exp_basic_06[8])
    check(s1.format({'form': 'value'}), exp_basic_06[9])
    check(s1.flattened().format({'form': 'value'}), exp_basic_06[10])
    check(s1.format(rules={'#c': {'form': 'value'}}), exp_basic_06[11])

    check(s1.format({}, {
        '#c': {'form': 'value'},
        '@term1': {
            'show-zeros': True
        },
    }), exp_basic_06[12])

    s2 = s1.flattened()
    check(s2.format({
        'show-zeros': True
    }, {
        '#c': {'form': 'value'},
    }), exp_basic_06[13])

    s3 = -a + b
    check(s3.format(), exp_basic_06[14])

    s4 = a + (-b)
    check(s4.format(), exp_basic_06[15])

    s5 = a + b + el + d
    s5.flatten()
    check(s5.format(), exp_basic_06[16])


exp_basic_07 = [
    r"a_{0} + a_{1} + a_{2} + a_{3} + a_{4}",
    r"a_{0} + a_{1} + a_{2} + \cdots + a_{4}",
]
def test_basic_07():
    a = FormalString('a')
    i = FormalVariable('i')
    ai = a._(i)
    #s = ai + ai + ai + ai + ai
    s0 = sum([ai]*5, FormalSum())
    s0.flatten()

    s1 = ai + ai + ai + ooo + ai
    s1.flatten()

    # print(s.format(rules={'i':7}))

    check(s0.format(rules={
        r'@term[i] #i': lambda i: i
    }), exp_basic_07[0])

    check(s1.format(rules={
        r'@term[i] #i': lambda i: i
    }), exp_basic_07[1])


exp_basic_08 = [
    r"a_{0} + a_{1} + \cdots + a_{p - 2}",
    r"a_{0} + a_{1} + \cdots + a_{1}",
    r"a_{0} + a_{1} + \cdots + a_{13 - 2}",
    r"a_{0} + a_{1} + \cdots + a_{11}",
    r"f: p \mapsto a_{0} + a_{1} + \cdots + a_{p - 2}",
    r"\sum_{i = 0}^{p - 2} a_{i}",
]
def test_basic_08():
    a = FormalString('a')
    i = FormalVariable('i')
    p = FormalInteger(13, 'p')
    s = FormalRangeSum([0, 1, ..., p - 2], a._(i), i)

    f = FormalMapping(FormalString('f'), None, None, [p], s)

    check(s.format(), exp_basic_08[0])

    check(s.format({}, {
        '@term3-sub': 1
    }), exp_basic_08[1])

    check(s.format({}, {
        '#p': {'form': 'value'}
    }), exp_basic_08[2])

    check(s.format({}, {
        '@term3-sub-subst': {'form': 'value'}
    }), exp_basic_08[3])

    check(f.format({
        'form': 'name-mapsto'
    }), exp_basic_08[4])

    check(s.format({
        'range': 'bind'
    }), exp_basic_08[5])


exp_basic_09 = [
    r"['2', '3', '5', '7']",
    r"['0', '1', '2', '3', '1', '2', '3', '4', '2', '3', '4', '5', '3', '4', '5', '6']",
    r"['1', '2', '3', '4', '1', '4', '9', '16', '1', '8', '27', '64', '1', '16', '81', '256']",
]
def test_basic_09():
    i = FormalVariable('i')
    j = FormalVariable('j')

    ai = FormalLookup([i], [2, 3, 5, 7])
    V = []
    for ival in range(4):
        V.append(ai.format(rules={
            '#i': ival
        }))
    check(V, exp_basic_09[0])

    sij = FormalLookup([i, j], lambda i, j: i + j)
    V = []
    for ival in range(4):
        for jval in range(4):
            V.append(sij.format(rules={
                '#i': ival,
                '#j': jval,
            }))
    check(V, exp_basic_09[1])

    tij = FormalLookup([i, j], [
        [1, 2, 3, 4],
        [1, 4, 9, 16],
        [1, 8, 27, 64],
        [1, 16, 81, 256],
    ])
    V = []
    for ival in range(4):
        for jval in range(4):
            V.append(tij.format(rules={
                '#i': ival,
                '#j': jval,
            }))
    check(V, exp_basic_09[2])


exp_basic_10 = [
    r"1",
    r"a^{0}",
    r"a",
    r"a^{1}",
    r"a^{2}",
    r"a^{2}",
]
def test_basic_10():
    a = FormalString('a')
    check((a ** 0).format(), exp_basic_10[0])
    check((a.sup(0)).format(), exp_basic_10[1])
    check((a ** 1).format(), exp_basic_10[2])
    check((a.sup(1)).format(), exp_basic_10[3])
    check((a ** 2).format(), exp_basic_10[4])
    check((a.sup(2)).format(), exp_basic_10[5])


exp_basic_11 = [
    r"\left(\left(a b\right) c\right) d",
    r"a b c d",
    r"c a d b",
    r"-b d",
    r"0",
    r"\left(a d\right) \left(c b\right)",
    r"210",
    r"0",
    r"-2 \cdot 0",
    r"-1 \cdot 2 \cdot 1 \cdot 0",
    r"(-1)(-2)(-1)(0)",
]
def test_basic_11():
    a = FormalVariable('a')
    b = FormalVariable('b')
    c = FormalVariable('c')
    d = FormalVariable('d')

    pr0 = a * b * c * d
    pr1 = pr0.flattened()

    check(pr0.format(), exp_basic_11[0])

    check(pr1.format(), exp_basic_11[1])
    check(pr1.permuted([2, 0, 3, 1]).format(), exp_basic_11[2])
    check(pr1.format(rules={
        '#a': 1,
        '#c': -1
    }), exp_basic_11[3])
    check(pr1.format(rules={
        '#d': 0
    }), exp_basic_11[4])

    pr2 = pr1.split([
        [0, 3], [2, 1]
    ])
    check(pr2.format(), exp_basic_11[5])

    print()

    check(pr1.format({
        'form': 'value'
    }, {
        '#a': 2, '#b': 3, '#c': 5, '#d': 7
    }), exp_basic_11[6])

    check(pr1.format({
        'mult-symb': 'dot',
    }, {
        '#a': -1, '#b': -2, '#c': -1, '#d': 0
    }), exp_basic_11[7])

    check(pr1.format({
        'mult-symb': 'dot',
        'collapse-zero': False
    }, {
        '#a': -1, '#b': -2, '#c': -1, '#d': 0
    }), exp_basic_11[8])

    check(pr1.format({
        'mult-symb': 'dot',
        'collapse-zero': False,
        'show-unity': True
    }, {
        '#a': -1, '#b': -2, '#c': -1, '#d': 0
    }), exp_basic_11[9])

    check(pr1.format({
        'mult-symb': 'paren',
        'collapse-zero': False,
        'show-unity': True,
        'signs': 'auto'
    }, {
        '#a': -1, '#b': -2, '#c': -1, '#d': 0
    }), exp_basic_11[10])


exp_basic_12 = [
    r"a_{0} + a_{1} \zeta + \cdots + a_{p - 2} \zeta^{p - 2}",
    r"$\left\{ a_{0} + a_{1} \zeta + \cdots + a_{p - 2} \zeta^{p - 2} \right\}$ where $a_i \in \mathbb{Q}$",
    r"the set of all $a_{0} + a_{1} \zeta + \cdots + a_{p - 2} \zeta^{p - 2}$ such that $a_i \in \mathbb{Q}$"
]
def test_basic_12():
    a = FormalString('a')
    i = FormalVariable('i')
    zeta = FormalVariable(r'\zeta', 'zeta')
    p = FormalInteger(11, 'p')

    alpha = FormalRangeSum([0, 1, ..., p-2], a._(i)*zeta**i, i)
    check(alpha.format(), exp_basic_12[0])

    K = FormalSet(alpha, condition=FormalString(r'a_i \in \mathbb{Q}'), name="K")
    check(K.format({
        'form': 'symbolic',
        'cond': 'where',
    }), exp_basic_12[1])

    check(K.format({
        'form': 'verbal'
    }), exp_basic_12[2])


exp_basic_13 = [
    r"\frac{1 - \zeta^{g}}{2}",
    r"\frac{1}{1 - \zeta^{g}}",
    r"\frac{\left(1 - \zeta^{g}\right)}{\left(1 - \zeta\right)}",
    r"\frac{\left(1 - \zeta^{6}\right)}{\left(1 - \zeta\right)}",
    r"\frac{\left(1 - \zeta^{3 \cdot 5}\right)}{\left(1 - \zeta\right)}",
    r"\frac{\left(1 - \zeta^{3 \cdot 5}\right)}{\left(1 - \zeta\right)}",
    r"\frac{\left(1 - \zeta^{3 \cdot 5}\right)}{\left(1 - \zeta\right)}",
]
def test_basic_13():
    zeta = FormalVariable(r'\zeta', 'zeta')
    g = FormalVariable('g')

    quo = (1 - zeta**g) / (1 - zeta)
    quo0 = (1 - zeta ** g) / 2
    quo1 = 1 / (1 - zeta ** g)

    check(quo0.format(), exp_basic_13[0])
    check(quo1.format(), exp_basic_13[1])

    check(quo.format({
        'brackets-top': 'round',
        'brackets-bot': 'round'
    }), exp_basic_13[2])

    check(quo.format({
        'brackets-top': 'round',
        'brackets-bot': 'round'
    }, {
        '#g': 6
    }), exp_basic_13[3])

    check(quo.format({
        'brackets-top': 'round',
        'brackets-bot': 'round'
    }, {
        '#g': FormalProduct([3, 5]),
        # We can set the mode for the thing we're substituting for `g`
        # by using its address in the syntax tree:
        '@top-term1-power-subst': {
            'mult-symb': 'dot'
        }
    }), exp_basic_13[4])

    check(quo.format({
        'brackets-top': 'round',
        'brackets-bot': 'round'
    }, {
        '#g': FormalProduct([3, 5]),
        # But it is easier to find it this way:
        '#g @subst': {
            'mult-symb': 'dot'
        }
    }), exp_basic_13[5])

    check(quo.format({
        'brackets-top': 'round',
        'brackets-bot': 'round'
    }, {
        # Another way to do it is to format the thing _before_ substituting it;
        # i.e. to just substitute a string.
        '#g': FormalProduct([3, 5]).format({
            'mult-symb': 'dot'
        })
    }), exp_basic_13[6])


def test_basic_14():
    q = FormalQuotient(3, 7)
    check(q.format({'form': 'value'}), "3/7")

    a = FormalInteger(0)
    b = FormalInteger(1)
    c = FormalInteger(2)
    d = FormalInteger(3)

    q0 = a / d
    check(q0.format(), '0')

    q1 = c / b
    check(q1.format(), '2')

    q2 = (-b) / d
    check(q2.format(), r'-\frac{1}{3}')

    q3 = d / (-b)
    check(q3.format(), '-3')

    q4 = (-c) / (-d)
    check(q4.format(), r'\frac{2}{3}')


exp_basic_15 = [
    r"\left\{ 0, 1, 2, \ldots, q - 1 \right\}",
    r"the set $\left\{ 0, 1, 2, \ldots, q - 1 \right\}$",
]
def test_basic_15():
    q = FormalInteger(11, 'q')
    S = FormalSet([0, 1, 2, ooo, q - 1], name='S')
    t = S.format({
        'form': 'symbolic'
    })
    check(t, exp_basic_15[0])
    check('the set ' + t, exp_basic_15[1])

    E = FormalSet([])
    check(E.format({
        'form': 'verbal'
    }), 'the empty set')
    check(E.format({
        'form': 'symbolic'
    }), r'\varnothing')


def test_basic_16():
    x = FormalVariable('x')
    y = FormalVariable('y')
    f = FormalMapping('f', args=[x, y], value_form=x**2 + y**2)
    s = f.format({
        'form': 'name-args'
    })
    check(s, 'f(x, y)')
    check('the function ' + s, 'the function $f(x, y)$')
    check(f.format({
        'form': 'name-mapsto'
    }), r'f: x, y \mapsto x^{2} + y^{2}')


def test_basic_17():
    x = FormalVariable('x')
    y = FormalVariable('y')
    a = FormalInteger(2)
    b = FormalInteger(-3)
    c = FormalInteger(-5)

    p0 = x * a * y * b * c
    p0.flatten()
    check(p0.format(), r'2 \cdot 3 \cdot 5 x y')

    p1 = x * a * b * y
    p1.flatten()
    check(p1.format(), r'-2 \cdot 3 x y')

    check(p1.format({
        'signs': 'auto'
    }), r'2 \cdot -3 x y')

    check(p1.format({
        'numerals': 'dot'
    }), r'-x \cdot 2 \cdot 3 y')

    check(p1.format({
        'numerals': 'none'
    }), '-x 2 3 y')


def test_basic_18():
    x = FormalVariable('x')
    a = FormalInteger(2, 'a')
    check(x.format(a), 'a')
    check(x.format(a, {
        '#x @subst': {
            'form': 'value'
        }
    }), '2')
    check(x.format({
        'subst': a,
        'form': 'value'
    }), '2')


def test_basic_19():
    a = FormalString('a')
    i = FormalVariable('i')
    p = FormalInteger(13, 'p')

    su = FormalRangeSum([0, 1, ..., p - 2], a._(i), i)
    se = FormalRangeSet([0, 1, ..., p - 2], a._(i), i)
    check(su.format(), r'a_{0} + a_{1} + \cdots + a_{p - 2}')
    check(se.format(), r'\left\{ a_{0}, a_{1}, \ldots, a_{p - 2} \right\}')
    check(su.format({
        'range': 'bind'
    }), r'\sum_{i = 0}^{p - 2} a_{i}')
    check(se.format({
        'range': 'bind'
    }), r'\left\{ a_{i} : 0 \leq i \leq p - 2 \right\}')

    su = FormalRangeSum([0, 1, ..., p - 2, ...], a._(i), i)
    se = FormalRangeSet([0, 1, ..., p - 2, ...], a._(i), i)
    check(su.format(), r'a_{0} + a_{1} + \cdots + a_{p - 2} + \cdots')
    check(se.format(), r'\left\{ a_{0}, a_{1}, \ldots, a_{p - 2}, \ldots \right\}')
    check(su.format({
        'range': 'bind'
    }), r'\sum_{i = 0}^{\infty} a_{i}')
    check(se.format({
        'range': 'bind'
    }), r'\left\{ a_{i} : 0 \leq i < \infty \right\}')

    su = FormalRangeSum([..., -2, -1, 0, 1, 2, ...], a._(i), i)
    se = FormalRangeSet([..., -2, -1, 0, 1, 2, ...], a._(i), i)
    check(su.format(), r'\cdots + a_{-2} + a_{-1} + a_{0} + a_{1} + a_{2} + \cdots')
    check(se.format(), r'\left\{ \ldots, a_{-2}, a_{-1}, a_{0}, a_{1}, a_{2}, \ldots \right\}')
    check(su.format({
        'range': 'bind'
    }), r'\sum_{i = -\infty}^{\infty} a_{i}')
    check(se.format({
        'range': 'bind'
    }), r'\left\{ a_{i} : -\infty < i < \infty \right\}')

    su = FormalRangeSum([2, -3, 5, -7])
    se = FormalRangeSet([2, -3, 5, -7])
    check(su.format(), r'2 - 3 + 5 - 7')
    check(se.format(), r'\left\{ 2, -3, 5, -7 \right\}')

    S = FormalSet([], name="S")
    su = FormalRangeSum(None, a._(i), i, cond = i |in_| S)
    se = FormalRangeSet(None, a._(i), i, cond = i |in_| S)
    check(su.format({
        'range': 'bind'
    }, {
        "#S": {
            'form': 'name'
        }
    }), r'\sum_{i \in S} a_{i}')
    check(se.format({
        'range': 'bind'
    }, {
        "#S": {
            'form': 'name'
        }
    }), r'\left\{ a_{i} : i \in S \right\}')


def test_basic_20():
    p = FormalVariable('p')
    f = (1 - 1/p)**(-1)

    # When no range and no condition given, then just put the bound var itself as subscript.
    pr0 = FormalRangeProd([], gen_form=f, bound_var=p)
    check(pr0.format(), r'\prod_{p} \left(1 - \frac{1}{p}\right)^{-1}')

    # However, do not do the same in the case of a set. In other words, we do
    # NOT want to get r'\left\{ \left(1 - \frac{1}{p}\right)^{-1} : p \right\}.
    S0 = FormalRangeSet([], gen_form=f, bound_var=p)
    check(S0.format(), r'\left\{ \left(1 - \frac{1}{p}\right)^{-1} \right\}')

    check(pr0.format(rules={
        '@form': {
            'negative': 'frac'
        }
    }), r'\prod_{p} \frac{1}{1 - \frac{1}{p}}')

    S = [2, 3, 5, 7]
    pr1 = FormalRangeProd(S, gen_form=f, bound_var=p)
    check(pr1.format(), r'\left(1 - \frac{1}{2}\right)^{-1} \left(1 - \frac{1}{3}\right)^{-1} \left(1 - \frac{1}{5}\right)^{-1} \left(1 - \frac{1}{7}\right)^{-1}')

    # Here the |in_| coerces the python list S into a FormalSet automatically:
    pr2 = FormalRangeProd([], gen_form=f, bound_var=p, cond=p |in_| S)
    check(pr2.format(), r'\prod_{p \in \left\{ 2, 3, 5, 7 \right\}} \left(1 - \frac{1}{p}\right)^{-1}')



# --------------------------------------------------------------------------- #
#   Stylgebra                                                                 #
#                                                                             #
#   Copyright (c) 2020-2023 Steve Kieffer                                     #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at                                   #
#                                                                             #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #

from stylgebra.basic import *


class InvertibleInfix:

    def __init__(self, function, valence=True):
        self.function = function
        self.valence = valence

    def __ror__(self, L):
        if isinstance(L, FormalInfixRelation):
            f = lambda R: L.toChain().extended([self, R])
        elif isinstance(L, FormalRelnChain):
            f = lambda R: L.extended([self, R])
        else:
            f = lambda R: self.function(L, R, valence=self.valence)
        return InvertibleInfix(f, self.valence)

    def __or__(self, R):
        return self.function(R)

    def __invert__(self):
        return InvertibleInfix(self.function, valence=not self.valence)


class FormalRelation(Formal):

    pass


class FormalBinaryRelation(FormalRelation):

    pass


class FormalInfixRelation(FormalBinaryRelation):

    def __init__(self, left, right, valence=True, name=None, id=None):
        super().__init__(name=name, id=id)
        self.left = wrap(left)
        self.right = wrap(right)
        self.valence = valence
        self.symbol = None
        self.nsymbol = None
        self.phrase = None
        self.nphrase = None
        # The "bar op" should be a reference to the InvertibleInfix instance
        # we define for each subclass of FormalInfixRelation. It is so called
        # because it is an operator used by surrounding it with vert bars: |.|
        self.barOp = None

    def toChain(self):
        return FormalRelnChain([self.left, self.barOp, self.right])

    @withmode
    def format(self, mode=None, rules=None, path=None, modifier=None):
        """
        options:

            omit-left: boolean
                Default: False
                If True, omit the lefthand term, but do prepend a space.
                This is useful for relation chains.

        subpaths:

            left: self.left
            right: self.right
        """
        form = mode.get('form')
        if form is None:
            if (self.valence and self.symbol is not None) or (not self.valence and self.nsymbol is not None):
                form = FormOpt.SYMBOLIC
            elif (self.valence and self.phrase is not None) or (not self.valence and self.nphrase is not None):
                form = FormOpt.VERBAL
            else:
                form = FormOpt.NAME

        if form == FormOpt.NAME:
            return MM(self.name())
        elif form == FormOpt.VALUE:
            # What would this mean? Decide whether the relation is true or false?
            raise NotImplementedError()

        left = '' if mode.get('omit-left') else self.fwd(self.left, 'left', rules, path)
        right = self.fwd(self.right, 'right', rules, path)

        if form == FormOpt.SYMBOLIC:
            return self.symbolic(mode, left, right)
        else:
            assert form == FormOpt.VERBAL
            return self.verbal(mode, left, right)

    def symbolic(self, mode, left, right):
        symb = self.symbol if self.valence else self.nsymbol
        return left + MM(f' {symb} ') + right

    def verbal(self, mode, left, right):
        phrase = self.phrase if self.valence else self.nphrase
        return left + f' {phrase} ' + right


class FormalSetMembershipRelation(FormalInfixRelation):

    def __init__(self, left, right, valence=True, name=None, id=None):
        if isinstance(right, list) or isinstance(right, set):
            # You can get away with passing a python list or set;
            # we will convert it into a FormalSet for you.
            from stylgebra.expressions import FormalSet
            right = FormalSet(elements=list(right))
        super().__init__(left, right, valence, name, id)
        self.symbol = r'\in'
        self.nsymbol = r'\not\in'
        self.phrase = 'is an element of'  # what about 'belongs to'? Should it be configurable?
        self.nphrase = 'is not an element of'
        self.barOp = in_


in_ = InvertibleInfix(FormalSetMembershipRelation)


class FormalLeqRelation(FormalInfixRelation):

    def __init__(self, left, right, valence=True, name=None, id=None):
        super().__init__(left, right, valence, name, id)
        self.symbol = r'\leq'
        self.nsymbol = r'\not\leq'
        self.phrase = 'is less than or equal to'
        self.nphrase = 'is not less than or equal to'
        self.barOp = leq


leq = InvertibleInfix(FormalLeqRelation)


class FormalLtRelation(FormalInfixRelation):

    def __init__(self, left, right, valence=True, name=None, id=None):
        super().__init__(left, right, valence, name, id)
        self.symbol = r'<'
        self.nsymbol = r'\nless'
        self.phrase = 'is less than'
        self.nphrase = 'is not less than'
        self.barOp = lt


lt = InvertibleInfix(FormalLtRelation)


class FormalRelnChain(Formal):
    """
    Represents a chain of terms and infix relations.
    The relations need not all be the same.
    """

    def __init__(self, chain):
        """
        :param chain: list of alternating formal terms and FormalInfixRelations,
            starting and ending with terms.
        """
        super().__init__()
        self.chain = chain

    def extend(self, ext):
        """
        Extend this chain in-place.
        """
        self.chain.extend(ext)

    def extended(self, ext):
        """
        Return a new instance, whose chain equals this one plus an extension.
        """
        return FormalRelnChain(self.chain + ext)

    @withmode
    def format(self, mode=None, rules=None, path=None, modifier=None):
        """
        subpaths:

            left: the leftmost term
            reln%d % i: the ith Infix relation for i from 1 to n - 1 when we have n terms.
        """
        chain = self.chain
        m = len(chain)
        assert m % 2 == 1
        n = (m - 1) // 2
        assert n > 0
        left = chain[0]
        s = self.fwd(left, 'left', rules, path)
        mod = {
            'omit-left': True
        }
        for i in range(n):
            reln, right = chain[2*i+1:2*i+3]
            ir = left |reln| right
            subpath = 'reln%d' % i
            s += self.fwd(ir, subpath, rules, path, modifier=mod)
            left = right
        return s

##############################################################################
# Tests


def try00():
    from stylgebra.expressions import FormalVariable
    a = FormalVariable('a')
    b = FormalVariable('b')
    c = FormalVariable('c')
    d = FormalVariable('d')
    rc = a |lt| b |leq| c |in_| d
    print(rc.format())


if __name__ == "__main__":
    try00()



# --------------------------------------------------------------------------- #
#   Stylgebra                                                                 #
#                                                                             #
#   Copyright (c) 2020-2023 Steve Kieffer                                     #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at                                   #
#                                                                             #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #

from sympy import primitive_root, totient, igcd, isprime, FF

from stylgebra.basic import *
from stylgebra.excep import FormalExcep


class FormalIntResidue(Formal):

    def __init__(self, r, m, name=None, id=None):
        super().__init__(name=name, id=id)
        if not isprime(m):
            raise FormalExcep(f'FormalIntResidue not yet supported for non-prime modulus {m}')
        self.modulus = m
        self.residue = r
        self.ring = FF(m)
        self.value(self.ring(r))

    @withmode
    def format(self, mode=None, rules=None, path=None, modifier=None):
        form = mode.get('form')
        if form is None:
            if self.name() is not None:
                form = FormOpt.NAME
            else:
                form = FormOpt.VALUE

        if form == FormOpt.NAME:
            return MM(self.name())
        elif form == FormOpt.VALUE:
            #return MM(self.value())
            return MM(int(self.value()))


def pick_primitive_root(modulus, auto_power=0):
    """
    You can easily get a primitive root r for a given modulus m using
    SymPy's `primitive_root` function. But you always get the same one.
    This function provides a convenient way to get the others.

    If g is the generator returned by `primitive_root`, and if r is
    the list of all units mod phi(modulus), then if you pass auto_power = k
    you will get the r[k] power of g.

    Example:
        If modulus == 13, then g = primitive_root(13) == 2.
        The available powers are [1, 5, 7, 11] (units mod 12 == 13 - 1).

        auto_power      return
        0               2^1 == 2
        1               2^5 == 6
        2               2^7 == 11
        3               2^11 == 7
    """
    g = primitive_root(modulus)
    if auto_power > 0:
        t = totient(modulus)
        s = totient(t)
        a = auto_power % s
        u = 1
        while a > 0:
            if igcd(u, t) == 1:
                a -= 1
        g = g**u % modulus
    return g



# --------------------------------------------------------------------------- #
#   Stylgebra                                                                 #
#                                                                             #
#   Copyright (c) 2020-2023 Steve Kieffer                                     #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at                                   #
#                                                                             #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #


import fractions

from sympy import isprime

from stylgebra.basic import *


class StrFormatArgs:
    """
    Instances of this class may serve as the `mode` argument when
    formatting a `FormalString`.
    """

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs


class FormalString(Formal):

    def __init__(self, s, mathmode=True):
        super().__init__()
        self.mathmode = mathmode
        self.s = MM(s) if mathmode else s

    @withmode
    def format(self, mode=None, rules=None, path=None, modifier=None):
        if isinstance(mode, StrFormatArgs):
            s = self.s.format(*mode.args, **mode.kwargs)
            return MM(s) if self.mathmode else s
        else:
            return self.s


class FormalVariable(Formal):

    def __init__(self, name, id=None):
        mode = name
        super().__init__(name=name, id=id, mode=mode)

    @withmode
    def format(self, mode=None, rules=None, path=None, modifier=None):
        r"""
        In the FormalVariable class, the mode plays an unusual role, in that
        it is used in order to substitute another object for the variable.

        Furthermore, if the substituted object is a Formal instance, then
        the variable takes on that object's value as its own; otherwise the
        variable takes on the object itself as its value.

        There are two ways to make the substitution:

        (1) The simplest way is to just pass the desired substitution itself as
        the mode.

        Then the substituted object can be addressed for formatting via the role
        of 'subst' relative to the variable. For example:

            i = FormalVariable('i')
            p = FormalInteger(7, 'p')

            i.format(p, {
                '#i @subst': {
                    'form': 'value'
                }
            })

        substitutes integer p for variable i, and then uses an ordinary rule
        dictionary to ask that p be printed by value.

        (2) Alternatively, you may pass a mode dict of the form

            {
                'subst': the thing to substitute,
                other key-value pairs to format the substituted thing...
            }

        When you use this format, the value under the key 'subst' is substituted for
        the variable, while the remainder of the mode dict is used as the _modifier_
        for the subst's mode. (Effectively, this means it will _be_ the subst's mode,
        unless you also address it directly via `@subst`, in which case the modifier
        will override.)

        Thus, our first example could also have been done this way,

            i.format({
                'subst': p,
                'form': 'value'
            })

        instead.
        """
        if isinstance(mode, dict) and 'subst' in mode:
            subst = mode['subst']
            subst_modifier = mode.copy()
            del subst_modifier['subst']
        else:
            subst = mode
            subst_modifier = None

        if isinstance(subst, Formal):
            self.value(subst.value())
        else:
            self.value(subst)

        return self.tryfwd(subst, 'subst', rules, path, modifier=subst_modifier, mathmode=True)


class FormalLookup(Formal):
    """
    This class can be thought of as a counterpart to the FormalMapping class.

    While the latter is meant to support writing down functions the way we typically
    do in mathematics, but is relatively weak in the types of function it can
    express via its `value_form` argument (which must be a Formal instance),
    the FormalLookup class is instead intended to support things like setting
    numerical coefficients for a polynomial, and is powerful in that you may
    pass any computable function, written in Python, as its `lookup`.

    Like the FormalMapping, however, the FormalLookup does read its inputs
    from Formal arguments.
    """

    def __init__(self, args, lookup, name=None, id=None):
        """
        args: a list of Formal objects (or a single Formal object), representing the
            argument(s) of the function
        lookup: either a callable, or a(n) (nested set of) array(s)/dict(s), in which to
            look up a value, based on the current values of the args. If callable, arity
            should match length of args list.
        """
        super().__init__(name=name, id=id)
        self.args = args if isinstance(args, list) else [args]
        self.lookup = lookup

    @withmode
    def format(self, mode=None, rules=None, path=None, modifier=None):
        """
        We make no use of our own mode, but our computed value may be addressed
        under the path extension `-value`, if it requires its own formatting.
        (Compare FormalVariable's `-subst` path extension.)
        """
        # Typically there may be FormalVariables (nested) among our args, and these
        # set their value during their format call, so we do need to recurse, but we
        # do not need the return values.
        for i, arg in enumerate(self.args):
            self.fwd(arg, 'arg%d' % i, rules, path)
        # Now we can compute our value, based on theirs.
        f = self.lookup
        if callable(f):
            # If f is callable, we expect its arity to match our number of args.
            v = f(*[a.value() for a in self.args])
        else:
            # Otherwise f should be indexable, and so should its values, and so
            # on, until we have exhausted all our args.
            for a in self.args:
                j = a.value()
                try:
                    f = f[j]
                except IndexError:
                    raise FormalExcep(f'Index {j} not in lookup: {f}')
            v = f
        self.value(v)
        return self.tryfwd(v, 'value', rules, path, mathmode=True)


class FormalInteger(Formal):

    def __init__(
            self, value=None, name=None, id=None,
            pos=None, odd=None, prime=None
    ):
        # Unlike so many cases where we do want built-in int's wrapped
        # as FormalIntegers, in the case of a FormalInteger itself, we
        # want its value to be a built-in int. So in case a FormalInteger
        # was passed, we dig down until we hit rock-bottom.
        while isinstance(value, FormalInteger):
            value = value.value()
        super().__init__(value=value, name=name, id=id)
        if value is None:
            self.pos = pos
            self.odd = odd
            self.prime = prime
        else:
            self.pos = value > 0
            self.odd = value % 2 == 1
            self.prime = isprime(value)

    @withmode
    def format(self, mode=None, rules=None, path=None, modifier=None):
        """
        mode options:

            form: FormOpt
                If undefined, behavior depends on what we have available:
                If we have a name, we use that; if not, then if we have a value,
                we use that; if not, then finally we use a verbal description.

            ordinal: boolean
                Default: False
                If form is _not_ 'verbal', and ordinal is True, then we will
                try to put the right ordinal suffix as a superscript.
        """

        form = mode.get('form')
        if form is None:
            if self.name() is not None:
                form = FormOpt.NAME
            elif self.value() is not None:
                form = FormOpt.VALUE
            else:
                form = FormOpt.VERBAL

        if form == FormOpt.NAME:
            s = MM(self.name())
        elif form in FormOpt.valsym:
            s = MM(self.value())
        else:
            assert form == FormOpt.VERBAL
            s = self.verbal()

        if mode.get('ordinal', False) and form != FormOpt.VERBAL:
            suffix = 'th'
            # The second condition here is so that we get 11th, 12th, 13th,
            # but 21st, 22nd, 23rd, etc.
            if s[-1:] in '123' and s[-2:-1] != '1':
                suffix = {
                    '1': 'st', '2': 'nd', '3': 'rd'
                }[s[-1:]]
            s = MM(r'%s^{\mathrm{%s}}' % (s, suffix))

        return s

    def verbal(self):
        adj_list = []

        if self.odd:
            adj_list.append('odd')
        elif self.odd is not None:
            adj_list.append('even')

        if self.pos:
            adj_list.append('positive')
        elif self.pos is not None:
            adj_list.append('negative')

        adjectives = ', '.join(adj_list)

        if self.prime:
            noun_phrase = 'prime'
        elif self.prime is not None:
            noun_phrase = 'composite integer'
        else:
            noun_phrase = 'integer'

        to_quantify = ''
        to_quantify += adjectives
        to_quantify += ' ' + noun_phrase
        to_quantify = to_quantify.strip()

        indef_art = 'a'

        if to_quantify[0] in 'aeiou':
            indef_art += 'n'

        d = indef_art + ' ' + to_quantify
        d = d.strip()

        return d


class FormalSubscripted(Formal):

    def __init__(self, base, sub, name=None, id=None):
        super().__init__(name=name, id=id)
        self.base = wrap(base)
        self.sub = wrap(sub)

    @withmode
    def format(self, mode=None, rules=None, path=None, modifier=None):
        """
        We make no use of our own mode.

        subpath names:
            'base': self.base
            'sub':  self.sub
        """
        b = self.fwd(self.base, 'base', rules, path)
        s = self.fwd(self.sub, 'sub', rules, path)
        return MM(r'%s_{%s}' % (b, s))


class FormalSuperscripted(Formal):
    """
    A FormalSuperscripted is different from a FormalPower in that it really
    _just_ means that something will be written as a superscript; it does not
    involve any of the special rules that come into play when the superscript
    means a power (like writing a**0 as 1 or a**1 as a). See FormalPower class.
    """

    def __init__(self, base, sup, name=None, id=None):
        super().__init__(name=name, id=id)
        self.base = wrap(base)
        self.sup = wrap(sup)

    @withmode
    def format(self, mode=None, rules=None, path=None, modifier=None):
        """
        We make no use of our own mode.

        subpath names:
            'base': self.base
            'sup':  self.sup
        """
        b = self.fwd(self.base, 'base', rules, path)
        s = self.fwd(self.sup, 'sup', rules, path)
        return MM(r'%s^{%s}' % (b, s))


class FormalPower(Formal):

    def __init__(self, base, power, name=None, id=None):
        super().__init__(name=name, id=id)
        self.base = wrap(base)
        self.power = wrap(power)

    def value(self):
        b = self.base.value()
        p = self.power.value()
        return b ** p

    @withmode
    def format(self, mode=None, rules=None, path=None, modifier=None):
        """
        mode options:

            form: FormOpt
                Default: symbolic

            show-zero: True/False
                Default: False
                If True, write b^0; if False, write 1.

            show-unity: True/False
                Default: False
                If True, write b^1; if False, write b.

            negative: [ 'sup', 'frac', 'inline', 'inline-paren' ]
                Default: 'sup'
                This controls how we handle negative powers.
                'sup': write as superscript, just like any other power.
                'frac': write as \frac{1}{x}
                'inline': write as 1/x
                'inline-paren': write as 1/(x)

            brackets-base: [ 'auto', 'none', 'round', 'square', 'curly' ]
                Default: 'auto'
                Control how the base is bracketed.
                'auto': we try to detect when brackets are needed, and apply them
                    accordingly.
                Otherwise as usual.

            brackets-power: [ 'auto', 'none', 'round', 'square', 'curly' ]
                Default: 'auto'
                Like brackets-base, but for the exponent.

        subpath names:
            'base':  self.base
            'power': self.power
        """
        b = self.fwd(self.base, 'base', rules, path)
        p = self.fwd(self.power, 'power', rules, path)

        # Note: Generally, when we want to return the value, we must only do
        # this _after_ forwarding the format operation to all our subterms.
        #
        # You might think this was unnecessary in such a case; however, FormalVariables
        # take on their values during their format call. So we need to cascade through
        # all the format calls below us, in order to ensure that every object has the
        # value it is supposed to have.
        #
        # You will see this pattern repeated in many classes in this module.

        form = mode.get('form', FormOpt.SYMBOLIC)
        if form == FormOpt.NAME:
            return MM(self.name())
        if form == FormOpt.VALUE:
            return MM(str(self.value()))

        outer_form = '%s'
        negative_policy = mode.get('negative', 'sup')
        if p[0] == '-' and negative_policy != 'sup':
            outer_form = {
                'frac': r'\frac{1}{%s}',
                'inline': '1/%s',
                'inline-paren': r'1/\left(%s\right)',
            }.get(negative_policy, outer_form)
            p = negate_string(p)

        show_zero = mode.get('show-zero', False)
        show_unity = mode.get('show-unity', False)

        base_brack = mode.get('brackets-base', 'auto')
        power_brack = mode.get('brackets-power', 'auto')

        def should_auto_bracket_base():
            if p == '1' and not show_unity:
                return False
            should_bracket = False
            should_bracket |= is_proper_sum(self.base)
            # FIXME:
            #  This next test really needs to be more nuanced. We want to check
            #  whether the quotient actually has appeared _as_ a quotient after
            #  formatting. Did its denominator disappear? If so then maybe don't
            #  need the brackets.
            should_bracket |= isinstance(self.base, FormalQuotient)
            return should_bracket

        if base_brack == 'auto':
            if should_auto_bracket_base():
                base_brack = 'round'

        if power_brack == 'auto' and is_proper_sum(self.power): power_brack = 'round'

        b = bracket_expression(base_brack, b)
        p = bracket_expression(power_brack, p)

        if p == '0' and not show_zero:
            return MM('1')
        elif p == '1' and not show_unity:
            return MM(outer_form % b)
        else:
            x = '%s^{%s}' % (b, p)
            return MM(outer_form % x)


class FormalSet(Formal):

    def __init__(self, elements, condition=None, name=None, id=None):
        """
        :param elements: either a list of formal elements, or a single formal element
        :param condition: optional formula satisfied by the elements
        """
        if not isinstance(elements, list):
            elements = [elements]
        elements = [wrap(e) for e in elements]
        self.elements = elements

        self.condition = condition
        Formal.__init__(self, name=name, id=id)

    @withmode
    def format(self, mode=None, rules=None, path=None, modifier=None):
        r"""
        mode options:

            form: FormOpt
                Default: symbolic

            empty: [ 'slashzero', 'braces' ]
                Say how empty set should be displayed.
                Default: 'slashzero'
                'slashzero': \varnothing
                'braces': \{\}

            cond: [ 'colon', 'vbar', 'where', 'with' ]
                Say how the condition should be displayed.
                Default: 'colon'
                'colon': after a colon (:) inside the braces
                'vbar': after a vertical bar (|) inside the braces
                'where': after the word "where" outside the braces
                'with': after the word "with" outside the braces

        subpaths:

            'elt%d' % i: self.elements[i]
            'cond': self.condition
        """

        elt_strings = [
            self.fwd(elt, 'elt%d' % i, rules, path, modifier)
            for i, elt in enumerate(self.elements)
        ]

        cond_string = self.fwd(self.condition, 'cond', rules, path) if self.condition is not None else None

        form = mode.get('form', FormOpt.SYMBOLIC)

        if form == FormOpt.NAME:
            return MM(self.name())
        elif form == FormOpt.VALUE:
            # Q: What could this actually mean? For now just leave it...
            return MM(self.value())
        elif form == FormOpt.VERBAL:
            return self.verbal(elt_strings, cond_string)

        assert form == FormOpt.SYMBOLIC

        n = len(self.elements)
        if n == 0:
            # Empty set
            empty_style = mode.get('empty', 'slashzero')
            if empty_style == 'braces':
                return MM(r'\{\}')
            else:
                return MM(r'\varnothing')

        elt_list = MM(', ').join(elt_strings)
        inner_cond = ''
        outer_cond = ''

        if cond_string is not None:
            cond_style = mode.get('cond', 'colon')
            if cond_style == 'colon':
                inner_cond = MM(' : ') + mbox(cond_string)
            elif cond_style == 'vbar':
                inner_cond = MM(r' \mid ') + mbox(cond_string)
            elif cond_style == 'where':
                outer_cond = ' where ' + cond_string
            elif cond_style == 'with':
                outer_cond = ' with ' + cond_string

        s = MM(r'\left\{ ') + elt_list + inner_cond + MM(r' \right\}') + outer_cond
        return s

    def verbal(self, elt_strings, cond_string=None):
        n = len(self.elements)
        if n == 0:
            s = 'the empty set'
        elif n > 1:
            s = 'the set containing ' + ', '.join(elt_strings)
        else:
            s = 'the set of all ' + elt_strings[0]
            if cond_string is not None:
                s += ' such that ' + cond_string
        return s


class FormalMapping(Formal):

    def __init__(self, name_form=None, domain=None, codomain=None, args=None, value_form=None, name=None, id=None):
        r"""
        name_form: a Formal object representing the name of the mapping
        domain: a Formal object representing the domain of the mapping
        codomain: a Formal object representing the codomain of the mapping
        args: a list of Formal objects, representing the argument(s) of the function
        value_form: a Formal object to represent the value of this mapping; generally,
            should involve the argument forms in one way or another.

        Example:
            sig = FormalString(r'\sigma')
            i = FormalVariable('i')
            p = FormalInteger('p', 11)
            F = CyclotomicField(p)
            zeta = FormalVariable('zeta', r'\zeta')

            sig_i = FormalMapping(sig._(i), F, F, [zeta], zeta**i, 'sig_i')

        """
        super().__init__(name=name, id=id)
        self.name_form = name_form
        self.domain = domain
        self.codomain = codomain
        self.args = args or []
        self.value_form = value_form

    @withmode
    def format(self, mode=None, rules=None, path=None, modifier=None):
        """
        mode options:

            form: [ 'name', 'name-args', 'value', 'map', 'mapsto', 'name-mapsto' ]
                Default: 'name'
                'name':         f
                'name-args':    f(args)
                'value':        value
                'map':          f: D --> C
                'mapsto':       args |--> value
                'name-mapsto':  f: args |--> value

        subpath names:
            'name':      self.name_form
            'arg%d' % i: self.args[i]
            'value':     self.value_form
            'domain':    self.domain
            'codomain':  self.codomain
        """
        form = mode.get('form', 'name')

        f, a, v, D, C = None, None, None, None, None
        if form not in ['value', 'mapsto']:
            f = self.fwd(self.name_form, 'name', rules, path)
        if form in ['name-args', 'mapsto', 'name-mapsto']:
            a = MM(', ').join([
                self.fwd(self.args[i], 'arg%d' % i, rules, path) for i in range(len(self.args))
            ])
        if form in ['value', 'mapsto', 'name-mapsto']:
            v = self.fwd(self.value_form, 'value', rules, path)
        if form == 'map':
            D = self.fwd(self.domain, 'domain', rules, path)
            C = self.fwd(self.codomain, 'codomain', rules, path)

        if form == 'name':
            return f
        elif form == 'name-args':
            return MM('%s(%s)' % (f, a))
        elif form == 'value':
            return v
        elif form == 'map':
            return MM(r'%s: %s \rightarrow %s' % (f, D, C))
        elif form == 'mapsto':
            return MM(r'%s \mapsto %s' % (a, v))
        elif form == 'name-mapsto':
            return MM(r'%s: %s \mapsto %s' % (f, a, v))


class FormalEllipsis(Formal):

    def __init__(self):
        super().__init__()

    @withmode
    def format(self, mode=None, rules=None, path=None, modifier=None):
        """
        mode options:

            style: [ 'l', 'c', 'v' ]
                Default: 'l'
                'l': ldots
                'c': cdots
                'v': vdots
        """
        s = mode.get('style', 'l')
        assert s in 'lcv'
        return MM('\\' + s + 'dots')


# We provide one general-purpose instance of the FormalEllipsis class,
# under a name that looks a bit like "...". The idea is that anyone can
# use this as a convenience when they need an ellipsis.
ooo = FormalEllipsis()


def is_ellipsis(thing):
    return thing == Ellipsis or isinstance(thing, FormalEllipsis)


class FormalInfinity(Formal):

    def __init__(self, sign=1, name=None, id=None):
        super().__init__(name=name, id=id)
        self.sign = sign

    @withmode
    def format(self, mode=None, rules=None, path=None, modifier=None):
        sign = MM('-') if self.sign == -1 else ''
        return sign + MM(r'\infty')


def negate_string(s):
    """
    Simple method of formally negating a string expression:
    If it begins with a '-', take that away; else add one.
    """
    if s[:1] == '-':
        s = MM(s[1:])
    else:
        s = MM('-') + s
    return s


class FormalSummand(Formal):

    def __init__(self, term, sign=1):
        super().__init__()
        self.term = wrap(term)
        self.sign = sign

    def value(self):
        return self.sign * self.term.value()

    @withmode
    def format(self, mode=None, rules=None, path=None, modifier=None):
        """
        Only one mode of operation. If sign is +1, we return our term's format result.
        If sign is -1, we try to ask our term to negate itself, and return the format
        result for the resulting term; if that cannot be done, we just tack a '-' sign
        onto the front of our term's format result, or try to take one away.

        submode names:
            'term': self.term
        """
        s = self.fwd(self.term, 'term', rules, path)
        if self.sign == -1:
            s = negate_string(s)
        return s

    def __neg__(self):
        return FormalSummand(self.term, -self.sign)

    def __add__(self, other):
        if isinstance(other, FormalSummand):
            return FormalSum([self, other])
        return NotImplemented

    def __sub__(self, other):
        if isinstance(other, FormalSummand):
            return FormalSum([self, -other])
        return NotImplemented


class MultiOp:
    """
    Mixin for "multi-operand" classes like FormalSum and FormalProduct.
    """

    def __init__(self, items):
        self.items = items

    def __len__(self):
        return len(self.items)

    def make_flat_items(self):
        flat = []
        for item in self.items:
            if isinstance(item, self.__class__):
                fi = item.make_flat_items()
                flat.extend(fi)
            else:
                flat.append(item)
        return flat

    def flatten(self):
        """
        Flatten this object's items array.
        Modifies this object in-place.
        See also `flattened` method.
        """
        self.items = self.make_flat_items()

    def flattened(self):
        """
        Return a new object whose items array is a flattened
        version of this one's. This object is unaltered.
        See also `flatten` method.
        """
        return self.__class__(self.make_flat_items())

    def permute(self, perm):
        """
        Permute the items according to a permutation given in "all targets"
        format, i.e. by listing the target to which each index gets mapped,
        using 0-based indices.

        For example, for the permutation

                    (0 1 2 3)
                    (0 2 1 3)

        that simply transposes indices 1 and 2, you would pass [0, 2, 1, 3].
        """
        self.items = [self.items[perm[i]] for i in range(len(perm))]

    def permuted(self, perm):
        """
        Like the `permute` method, only this time return a new object, instead
        of changing this one in-place.
        """
        items = [self.items[perm[i]] for i in range(len(perm))]
        return self.__class__(items)

    def split(self, parts, alt_cls=None):
        """
        Split this multi into a multi of multis.

        :param parts: list of list of indices. Each list says which operands
          you want in that group, and in what order. So you can also achieve
          a permutation at the same time.

        :param alt_cls: "alternate class". If given, this class will be used
          as the outer multi, which has all the others as its items.

        Example:
            pr1 = FormalProduct([a, b, c, d]).flattened()
            pr2 = pr1.split([
                [0, 3], [2, 1]
            ])

            pr1.format()
            # a b c d

            pr2.format()
            # (a d)(c b)
        """
        inner_cls = self.__class__
        outer_cls = alt_cls or self.__class__
        multis = [
            inner_cls([self.items[i] for i in part])
            for part in parts
        ]
        return outer_cls(multis)


class FormalSum(Formal, MultiOp):

    def __init__(self, terms=None, name=None, id=None):
        """
        :param terms: list of terms to be summed. If a term is not already a
          FormalSummand, we will wrap it in one. If you want a term to be subtracted
          rather than added, then you need to wrap it as a FormalSummand yourself,
          and set the sign to -1.
        """
        terms = terms or []
        # Ensure that everything is wrapped in a FormalSummand
        summands = [
            t if isinstance(t, FormalSummand) else FormalSummand(t)
            for t in terms
        ]
        Formal.__init__(self, name=name, id=id)
        MultiOp.__init__(self, summands)

    def summands(self):
        return self.items

    def value(self):
        return sum([s.value() for s in self.summands()])

    def make_flat_items(self):
        """
        Due to special structure with each term being wrapped
        in a FormalSummand, and the possible need to negate,
        we need to override the superclass method.
        """
        flat = []
        for summand in self.summands():
            term = summand.term
            if isinstance(term, FormalSum):
                fs = term.make_flat_items()
                if summand.sign == -1:
                    flat.extend([-s for s in fs])
                else:
                    flat.extend(fs)
            else:
                flat.append(summand)
        return flat

    @withmode
    def format(self, mode=None, rules=None, path=None, modifier=None):
        """
        mode options:

            form: FormOpt
                Default: symbolic

            brackets: [ 'none', 'round', 'square', 'curly' ]
                Default: 'none'
                Surround the entire expression with some kind of brackets.

            show-zeros: True/False
                Default: False
                If True, show terms equal to zero (0). If False, suppress these.

            flip-ops: True/False
                Default: True
                If True, flip operations +/- when the next term begins with a
                  minus sign '-'.

        subpath names:
            'term%d' % i: self.summands()[i].term
        """

        def make_modifier(term):
            if isinstance(term, FormalSum):
                # If the term is a sum, force it to be parenthesized.
                return {'brackets': 'round'}
            elif isinstance(term, FormalEllipsis):
                # Ellipses must be center-style
                return {'style': 'c'}
            else:
                return None

        term_strings = [
            self.fwd(summand.term, 'term%d' % i, rules, path, modifier=make_modifier(summand.term))
            for i, summand in enumerate(self.summands())
        ]

        form = mode.get('form', FormOpt.SYMBOLIC)
        if form == FormOpt.NAME:
            return MM(self.name())
        if form == FormOpt.VALUE:
            return MM(str(self.value()))

        s = ''
        first = True
        multiple_terms = False

        show_zeros = mode.get('show-zeros', False)
        flip_ops = mode.get('flip-ops', True)

        for f, summand in zip(term_strings, self.summands()):
            sign = summand.sign

            # Omit zeros if desired
            if f in ['0', '-0'] and not show_zeros: continue

            if first:
                first = False
                if sign == -1:
                    f = negate_string(f)
                s = f
            else:
                # Flip sign if appropriate
                if f[0] == '-' and flip_ops:
                    sign *= -1
                    f = MM(f[1:])

                op = {-1: '-', 1: "+"}[sign]
                s += MM(' %s %s' % (op, f))
                multiple_terms = True

        if s == '':
            # An empty sum equals zero.
            s = '0'

        bracket_style = mode.get('brackets', 'none')
        if multiple_terms and bracket_style != 'none':
            s = bracket_expression(bracket_style, s)

        return MM(s)


def is_proper_sum(obj):
    """
    Say whether a formal object appears to be a sum of at least two terms.
    This may be useful in judging whether an expression needs to be surrounded
    by brackets.
    """
    return isinstance(obj, FormalSum) and len(obj) >= 2


class FormalRangeBuilder(Formal):
    r"""
    Abstract base class for a variety of kinds of mathematical expressions, in which
    some "range" of terms or elements are assembled or combined in one way or another.

    In particular, this class supports styling options to spell out all the elements,
    or to instead express a condition on a bound variable.

    For example, this class will have a subclass to support the construction of sums,
    and one to support the construction of sets. Given the same initialization,

        rng = [0, 1, ..., p - 2]
        gen_form = a._{i}
        bound_var = i

    these subclasses could produce both spelled-out formats:

        sum:
            a_{0} + a_{1} + \ccc + a_{p - 2}

        set:
            \{ a_{0}, a_{1}, \ddd, a_{p - 2} \}

    or, with different styling options, bound formats:

        sum:
            \Sum_{i = 0}^{p - 2} a_i

        set:
            \{ a_i : 0 \leq i \leq p - 2 \}
    """

    def __init__(self, rng=None, gen_form=None, bound_var=None, cond=None, name=None, id=None):
        """
        rng: the "range", i.e. a list indicating the range of terms or elements
            involved in the construction. May include ellipses.
        gen_form: a Formal object representing the generic form of an element.
        bound_var: a FormalVariable representing the bound variable over which
            the range extends. Should occur at least once in the gen_form.

        If gen_form and bound_var are defined then the range is interpreted as
        a list of values that are to be substituted for the bound variable.
        If they are not both defined, then the range is interpreted as the list
        of desired terms or elements themselves.
        """
        super().__init__(name=name, id=id)
        self.range = rng or []
        if gen_form is None or bound_var is None:
            bound_var = FormalVariable('i')
            # Somewhat hacky: since we will be addressing var i with a selector that expects
            # to find it strictly _within_ the generic term, we can't just make the
            # latter equal to i; we have to put i within it.
            gen_form = FormalVariable('j')
            gen_form.mode(bound_var)
        self.gen_form = gen_form
        self.bound_var = bound_var
        self.cond = cond

        # Instead of overriding the make_subst_rules method, subclasses may opt
        # to simply define an item_subpath_name, and let the generic method do
        # the rest. For many (if not all) subclasses, this will suffice.
        # For example, the `FormalRangeSum` class need only set item_subpath_name
        # equal to `term`; the `FormalRangeSet` class need only set it equal
        # to `elt`.
        self.item_subpath_name = None

    def make_subst_rules(self, basepath):
        if self.item_subpath_name is None:
            raise NotImplementedError()
        if basepath: basepath += '-'
        rule = f'@{basepath}{self.item_subpath_name}[i] #{self.bound_var.id()}'
        return {
            rule: lambda i: self.range[i],
        }

    def interpolate_range(self, step_size):
        if len(self.range) == 2 and step_size != 0:
            # I suppose the endpts need to be FormalIntegers.
            # But our task is actually quite tricky since even if an endpt has
            # a numerical value, the question of whether we want an ellipsis is
            # really the question of how we want the endpt to be displayed, i.e.
            # by name or by value. So we need to do some kind of formatting lookahead
            # and check which mode is going to be applied (NAME or VALUE or other)
            # to the FormalInteger in question.
            ...  # TODO
            raise NotImplementedError("range interpolation feature not supported yet")
        else:
            return self.range[:]

    def build_subobject_for_list(self, items):
        """
        Subclasses must override.
        """
        raise NotImplementedError()

    def bound_format_with_range_bounds(self, a, b, mode, rules, path, modifier):
        """
        Subclasses must override.
        Produce the desired formatting where the condition on the bound variable
        is that it lie in the range from a to b.

        a: lower bound for range
        b: upper bound for range

        Other parameters are as in the `format` method.

        :return: string
        """
        raise NotImplementedError()

    def bound_format_with_generic_condition(self, cond, mode, rules, path, modifier):
        """
        Subclasses must override.
        Produce the desired formatting where the condition on the bound variable
        is a generic one.

        cond: the condition on the bound variable.

        Other parameters are as in the `format` method.

        :return: string
        """
        raise NotImplementedError()

    @withmode
    def format(self, mode=None, rules=None, path=None, modifier=None):
        r"""
        options:

            range: [ 'auto', 'list', 'bind' ]
                Default: auto
                list: write out all the elements, making substitutions from the range
                    into the gen_form through the bound_var.
                bind: express the gen_form as is, together with conditions on the bound_var.
                    Unless condition was explicitly given, assume it is boundedness by the
                    first and last elements of the range. Interpret the latter as -/+ \infty
                    if they are ellipses.
                auto: choose behavior based on length n of the range. If n >= 3 then operate
                    in list mode; if n == 2, then operate in bind mode if step option (see
                    below) is zero, otherwise list mode; if n == 1 operate in bind mode but
                    do not infer any condition from the range.

                In all cases, if the range is of length 0 then behavior is controlled by the
                subclass. E.g. empty sums should be '0', empty products should be '1', etc.

            step: int
                Default: 0
                step is only relevant if mode is 'list' and the range is length 2.
                Then step == 0 means do not fill in any additional items.
                Otherwise fill in additional items according to the step size.
                If either endpt of the range is non-numerical then use an ellipsis.

        """
        range_mode = mode.get('range', 'auto')
        step_size = mode.get('step', 0)
        try: step_size = int(step_size)
        except ValueError: raise FormalExcep(f'Malformed step size "{step_size}". Should be integer.')
        len_range = len(self.range)
        if range_mode == 'auto':
            if len_range >= 3:
                range_mode = 'list'
            elif step_size != 0:
                range_mode = 'list'
            else:
                range_mode = 'bind'

        if range_mode == 'list':
            #range_ = self.interpolate_range(step_size)
            range_ = self.range[:]
            rules = rules or {}
            path = path or ExpressionPath()
            basepath = path.rolepath()
            subst_rules = self.make_subst_rules(basepath)
            items = [
                ooo if is_ellipsis(a) else self.gen_form
                for a in range_
            ]
            subobject = self.build_subobject_for_list(items)
            assert isinstance(subobject, Formal)
            rules.update(subst_rules)
            return subobject.format(mode, rules, path, modifier)

        else:
            assert range_mode == 'bind'
            if self.cond is None and len_range >= 2:
                a = self.range[0]
                b = self.range[-1]
                a = FormalInfinity(sign=-1) if is_ellipsis(a) else wrap(a)
                b = FormalInfinity(sign=1) if is_ellipsis(b) else wrap(b)
                return self.bound_format_with_range_bounds(a, b, mode, rules, path, modifier)
            else:
                if self.cond is None and isinstance(self, FormalRangeOperator):
                    self.cond = self.bound_var
                return self.bound_format_with_generic_condition(self.cond, mode, rules, path, modifier)


class FormalRangeSet(FormalRangeBuilder):

    def __init__(self, rng=None, gen_form=None, bound_var=None, cond=None, name=None, id=None):
        super().__init__(rng, gen_form, bound_var, cond, name, id)
        self.item_subpath_name = 'elt'

    def build_subobject_for_list(self, items):
        return FormalSet(items, condition=self.cond)

    def bound_format_with_generic_condition(self, cond, mode, rules, path, modifier):
        S = FormalSet([self.gen_form], condition=cond)
        return S.format(mode, rules, path, modifier)

    def bound_format_with_range_bounds(self, a, b, mode, rules, path, modifier):
        from stylgebra.relations import lt, leq
        l_reln = lt if isinstance(a, FormalInfinity) else leq
        r_reln = lt if isinstance(b, FormalInfinity) else leq
        cond = a |l_reln| self.bound_var |r_reln| b
        return self.bound_format_with_generic_condition(cond, mode, rules, path, modifier)


class FormalRangeOperator(FormalRangeBuilder):
    """
    This abstract class unifies all those FormalRangeBuilders that work like
    Sum or Product, in the way they express bound format, be it for a range
    of indices, or a generic condition.
    """

    def __init__(self, rng=None, gen_form=None, bound_var=None, cond=None, name=None, id=None):
        super().__init__(rng, gen_form, bound_var, cond, name, id)
        # Subclasses must define their operator symbol.
        # E.g. r'\sum', r'\prod', etc.
        self.operator_symbol = None

    def build_subobject_for_list(self, items):
        """
        Subclasses must override.
        """
        raise NotImplementedError()

    def bound_format_with_range_bounds(self, a, b, mode, rules, path, modifier):
        """
        subpaths:
            form: self.gen_form
            var: self.bound_var
            lower: a
            upper: b
        """
        # TODO:
        #   Maybe use more sophisticated TeX?
        #   See: <https://tex.stackexchange.com/a/85072>
        #        <https://www.overleaf.com/learn/latex/Integrals,_sums_and_limits>
        if self.operator_symbol is None:
            raise NotImplementedError()
        form = self.fwd(self.gen_form, 'form', rules, path)
        var = self.fwd(self.bound_var, 'var', rules, path)
        lower = self.fwd(a, 'lower', rules, path)
        upper = self.fwd(b, 'upper', rules, path)
        op = MM(self.operator_symbol)
        subscript = MM('_{') + var + MM(' = ') + lower + MM('}')
        superscript = MM('^{') + upper + MM('} ')
        return op + subscript + superscript + form

    def bound_format_with_generic_condition(self, cond, mode, rules, path, modifier):
        """
        subpaths:
            form: self.gen_form
            cond: cond
        """
        if self.operator_symbol is None:
            raise NotImplementedError()
        form = self.fwd(self.gen_form, 'form', rules, path)
        cond = self.fwd(cond, 'cond', rules, path)
        op = MM(self.operator_symbol)
        subscript = MM('_{') + cond + MM('} ')
        return op + subscript + form


class FormalRangeSum(FormalRangeOperator):

    def __init__(self, rng=None, gen_form=None, bound_var=None, cond=None, name=None, id=None):
        super().__init__(rng, gen_form, bound_var, cond, name, id)
        self.operator_symbol = r'\sum'
        self.item_subpath_name = 'term'

    def build_subobject_for_list(self, items):
        return sum(items, FormalSum()).flattened()

    def generic_term(self):
        return self.gen_form


class FormalRangeProd(FormalRangeOperator):

    def __init__(self, rng=None, gen_form=None, bound_var=None, cond=None, name=None, id=None):
        super().__init__(rng, gen_form, bound_var, cond, name, id)
        self.operator_symbol = r'\prod'
        self.item_subpath_name = 'factor'

    def build_subobject_for_list(self, items):
        return FormalProduct(factors=items)


"""
A word on the implementation differences between addition & subtraction on the
one hand, and multiplication & division on the other:

This module is supposed to help us model _common_ practice in mathematical
expression, of the kind you see in textbooks, and on chalkboards. And, whereas
it is common to see long chains of mixed adding and subtracting,

    a + b - c + d - e - f

it is in no way common to see similar chains of mixed multiplying and dividing
in such contexts:

    a * b / c * d / e / f

This is why, while it seemed suitable to use a FormalSummand class, with a
positive and negative modality, in order to model the former (in combination
with the FormalSum class), we do _not_ plan to have an analogous "FormalFactor"
class.

Instead, a FormalProduct and FormalQuotient class should suffice to let us model
the common practice with multiplication & division, wherein we tend to have a
single quotient, with a product upstairs and a product downstairs, as in

    (a*b*d) / (c*e*f)

(If we were to do likewise with addition & subtraction then the sum above
would be gathered as (a + b + d) - (c + e + f). While certainly acceptable,
this is in no way required by convention.) 
"""


class FormalProduct(Formal, MultiOp):

    def __init__(self, factors=None, name=None, id=None):
        Formal.__init__(self, name=name, id=id)
        factors = [wrap(f) for f in factors]
        MultiOp.__init__(self, factors or [])

    def factors(self):
        return self.items

    def value(self):
        v = 1
        for f in self.factors():
            v *= f.value()
        return v

    @withmode
    def format(self, mode=None, rules=None, path=None, modifier=None):
        """
        mode options:

            form: FormOpt
                Default: symbolic

            brackets: [ 'none', 'round', 'square', 'curly' ]
                Default: 'none'
                Surround the entire expression with some kind of brackets.

            collapse-zero: True/False
                Default: True
                If True, and if any factor looks like zero, then return '0' for the
                  entire product.
                If False, include zeros just like any other factor.

            mult-symb: [ 'none', 'dot', 'x', 'paren' ]
                Default: 'none'
                Say what multiplication symbol you want to appear between (or around) factors.

            numerals: [ 'front-dot', 'dot', 'none' ]
                Default: 'front-dot'
                Say how to handle numeral factors.
                At present the set of factors regarded as "numerals" is quite limited: their
                    string representation has to parse as an integer. We may expand this in
                    future versions.
                'dot': if 'mult-symb' is 'none', then put a cdot before any numeral
                    (except for the leftmost factor in the entire product).
                'front-dot': like 'dot', plus bring numerals to the front (left side)
                    of the product.
                'none': no special processing for numerals.

            show-unity: True/False
                Default: False
                If True, show factors equal to unity (1). If False, suppress these.

            signs: [ 'gather', 'bracket', 'auto' ]
                Default: 'gather'
                Say how to treat factors whose expression begins with a minus sign '-'.
                'gather': Negations are counted up, and at most a single minus
                  sign will appear at the front of the whole product (according
                  to parity of the count), with all others suppressed.
                'bracket': All negated factors are bracketed.
                'auto': Negated factors are bracketed unless prefixed by some multiplication
                    symbol due to the `mult-symb` option or the `numerals` option.

        subpaths:
            'factor%d' % i: self.factors()[i]
        """

        def make_modifier(factor):
            if isinstance(factor, FormalSum):
                return {'brackets': 'round'}
            elif isinstance(factor, FormalProduct):
                return {'brackets': 'round'}
            elif isinstance(factor, FormalEllipsis):
                # Ellipses must be center-style
                return {'style': 'c'}
            else:
                return None

        factor_strings = [
            self.fwd(factor, 'factor%d' % i, rules, path, modifier=make_modifier(factor))
            for i, factor in enumerate(self.factors())
        ]

        form = mode.get('form', FormOpt.SYMBOLIC)
        if form == FormOpt.NAME:
            return MM(self.name())
        if form == FormOpt.VALUE:
            return MM(self.value())

        mult_symb_rule = mode.get('mult-symb', 'none')
        mult_symbol_str = MM({
            'none': ' ',
            'dot': r' \cdot ',
            'x': r' \times ',
            'paren': ')('
        }[mult_symb_rule])

        collapse_zero = mode.get('collapse-zero', True)
        show_unity = mode.get('show-unity', False)

        sign_rule = mode.get('signs', 'gather')
        gather_signs = sign_rule == 'gather'
        bracket_signs = sign_rule == 'bracket' or (sign_rule == 'auto' and mult_symbol_str == ' ')

        numeral_rule = mode.get('numerals', 'front-dot')
        gather_numerals = numeral_rule == 'front-dot'
        dot_numerals = (mult_symb_rule == 'none' and numeral_rule[-3:] == 'dot')

        pruned = []
        numerals = []
        num_neg = 0
        num_numerals = 0
        num_factors = 0
        for s in factor_strings:
            needs_brackets = False

            if s in ['0', '-0'] and collapse_zero:
                product = MM('0')
                break

            if s[0] == '-':
                if gather_signs:
                    num_neg += 1
                    s = MM(s[1:])
                elif bracket_signs:
                    needs_brackets = True

            if s != '1' or show_unity:

                treat_as_numeral = False
                if gather_numerals or dot_numerals:
                    try: int(s)
                    except ValueError: pass
                    else: treat_as_numeral = True

                if dot_numerals and treat_as_numeral:
                    if num_numerals > 0 or (num_factors > 0 and not gather_numerals):
                        s = MM(r'\cdot ') + s
                        needs_brackets = False

                if needs_brackets:
                    s = MM('(') + s + MM(')')

                if gather_numerals and treat_as_numeral:
                    numerals.append(s)
                    num_numerals += 1
                else:
                    pruned.append(s)

                num_factors += 1
        else:
            # We didn't break, meaning either we never encountered a zero, or
            # we want to show zeros like regular factors.
            if gather_numerals:
                pruned = numerals + pruned
            sign = MM('-') if num_neg % 2 == 1 else ''
            term = MM('1') if num_factors == 0 else mult_symbol_str.join(pruned)
            if mult_symbol_str == ')(':
                term = MM("(") + term + MM(")")
            product = sign + term

        bracket_style = mode.get('brackets', 'none')
        product = bracket_expression(bracket_style, product)

        return MM(product)


def formalQuoFromFrac(frac):
    """
    Turn an instance of python's fractions.Fraction class into a FormalQuotient.
    """
    return FormalQuotient(frac.numerator, frac.denominator)


class FormalQuotient(Formal):

    def __init__(self, top, bot, name=None, id=None):
        super().__init__(name=name, id=id)
        self.top = wrap(top)
        self.bot = wrap(bot)

    def value(self):
        vt = self.top.value()
        vb = self.bot.value()
        if isinstance(vt, int) and isinstance(vb, int):
            vq = fractions.Fraction(vt, vb)
        else:
            vq = vt / vb
        return vq

    @withmode
    def format(self, mode=None, rules=None, path=None, modifier=None):
        """
        mode options:

            form: FormOpt
                Default: symbolic

            inline: True/False
                Default: False
                If True, write as top/bot; if False, write as \frac{top}{bot}.

            brackets-top: [ 'none', 'round', 'square', 'curly' ]
                Default: 'none'
                Surround the entire top expression with some kind of brackets.

            brackets-bot: [ 'none', 'round', 'square', 'curly' ]
                Default: 'none'
                Surround the entire bottom expression with some kind of brackets.

            collapse-zero: True/False
                Default: True
                If True, and if the numerator looks like zero, then return '0' for the
                  entire quotient.

            sign-front: True/False
                Default: True
                If True, attempt to cancel signs in numerator and denominator, and bring
                    any remaining negation out in front of the expression.

            collapse-int: True/False
                Default: True
                If True, and if the denominator looks like unity, then just write the
                    numerator.

        subpaths:
            'top': self.top
            'bot': self.bot
        """
        t = self.fwd(self.top, 'top', rules, path)
        b = self.fwd(self.bot, 'bot', rules, path)

        form = mode.get('form', FormOpt.SYMBOLIC)
        if form == FormOpt.NAME:
            return MM(self.name())
        if form == FormOpt.VALUE:
            return MM(self.value())

        collapse_zero = mode.get('collapse-zero', True)
        sign_front = mode.get('sign-front', True)
        collapse_int = mode.get('collapse-int', True)

        if t in ['0', '-0'] and collapse_zero:
            return MM('0')

        num_neg = 0
        if sign_front:
            if t[0] == '-':
                num_neg += 1
                t = MM(t[1:])
            if b[0] == '-':
                num_neg += 1
                b = MM(b[1:])

        top_bracket_style = mode.get('brackets-top', 'none')
        bot_bracket_style = mode.get('brackets-bot', 'none')

        t = bracket_expression(top_bracket_style, t)
        b = bracket_expression(bot_bracket_style, b)

        if b == '1' and collapse_int:
            q = t
        elif mode.get('inline'):
            q = '%s/%s' % (t, b)
        else:
            q = r'\frac{%s}{%s}' % (t, b)

        if num_neg == 1:
            q = f'-{q}'

        return MM(q)

##############################################################################
# Tests

def try01():
    a = FormalInteger(1, 'a')
    b = FormalInteger(2, 'b')
    c = FormalInteger(0, 'c')

    s0 = a + b + c
    print(s0.format())

def try02():
    a = FormalVariable('a')
    b = FormalVariable('b')
    c = FormalVariable('c')
    d = FormalVariable('d')

    pr0 = a * b * c * d
    print(pr0.format())

def try03():
    end_pattern = 'cat-foo[i]bar[j]'
    paths = [
        'baz3-cat-foo17bar81',
        'cat-foo0bar1',
        'foo2bar3',
        'cat-foo5bar'
    ]
    for path in paths:
        print("="*50)
        print(path)
        R = path.split('-')
        A = [None] * len(R)
        ep = ExpressionPath(A, R)
        pm = PathMatch(end_pattern, ep)
        print(pm.matched)
        if pm.matched:
            print(pm.indices)
            print(pm.num_segments_matched)
            print(pm.rolepath_prefix)
            print(pm.exp_path_prefix)

def try04():
    zeta = FormalVariable(r'\zeta', 'zeta')
    g = FormalVariable('g')
    quo = (1 - zeta ** g) / (1 - zeta)
    s = quo.format({
        'brackets-top': 'round',
        'brackets-bot': 'round'
    }, {
        '#g': FormalProduct([3, 5]),
        # But it is easier to find it this way:
        '#g @subst': {
            'mult-symb': 'dot'
        }
    })
    print(s)

def try05():
    a = FormalString('a')
    i = FormalVariable('i')
    zeta = FormalVariable(r'\zeta', 'zeta')
    p = FormalInteger(11, 'p')
    alpha = FormalRangeSum([0, 1, ..., p - 2], a._(i) * zeta ** i, i)
    s = alpha.format()
    print(s)

def try06():
    b = FormalInteger(1)
    d = FormalInteger(3)
    q2 = (-b) / d
    s = q2.format()
    print(s)

def try07():
    x = FormalVariable('x')
    y = FormalVariable('y')
    a = FormalInteger(2)
    b = FormalInteger(-3)
    p1 = x * a * b * y
    p1.flatten()
    s = p1.format({
        'signs': 'auto'
    })
    print(s)

def try08():
    x = FormalVariable('x')
    a = FormalInteger(2, 'a')
    s1 = x.format(a, {
        '#x @subst': {
            'form': 'value'
        }
    })
    print(s1)
    s2 = x.format({
        'subst': a,
        'form': 'value'
    })
    print(s2)

def try09():
    from stylgebra.relations import in_
    p = FormalVariable('p')
    f = (1 - 1/p)**(-1)
    S = [2, 3, 5, 7]
    pr2 = FormalRangeProd([], gen_form=f, bound_var=p, cond=p |in_| S)
    s = pr2.format()
    print(s)

if __name__ == "__main__":
    try09()



# --------------------------------------------------------------------------- #
#   Stylgebra                                                                 #
#                                                                             #
#   Copyright (c) 2020-2023 Steve Kieffer                                     #
#                                                                             #
#   Licensed under the Apache License, Version 2.0 (the "License");           #
#   you may not use this file except in compliance with the License.          #
#   You may obtain a copy of the License at                                   #
#                                                                             #
#       http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                             #
#   Unless required by applicable law or agreed to in writing, software       #
#   distributed under the License is distributed on an "AS IS" BASIS,         #
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#   See the License for the specific language governing permissions and       #
#   limitations under the License.                                            #
# --------------------------------------------------------------------------- #

import re
import inspect

from sympy import Integer as SympyInteger
from sympy import Rational as SympyRational

from stylgebra.excep import FormalExcep


class MM(str):
    """
    Use this subclass of str to represent strings intended to go inside LaTeX
    math modes. It provides smart concatenation so that, if a:T means a is a string
    of type T, then we have:

            a:mm + b:mm   = ab:mm
            a:mm + b:str  = $a$b:str
            a:str + b:mm  = a$b$:str

    and of course

            a:str + b:str = ab:str

    is still true!
    """

    def __new__(cls, value):
        """
        We want `MM(a)` to work for many different types of argument `a`.
        If `a` is a str, then we represent that string in math mode.
        If `a` is another MM, we want to represent the same string it does.
        If `a` is anything else, we want to convert it into a pure string,
        and then represent that string in math mode.
        """
        # While the given value is not in the subclass hierarchy of `str`, or
        # it is but lies at or below `MM`, then convert it into a `str`.
        while isinstance(value, MM) or not isinstance(value, str):
            value = str(value)
        # Now we can make a plain string on that value.
        self = str.__new__(cls, value)
        # And store it as the plain string that we represent.
        self.plain_str = value
        return self

    def __str__(self):
        """
        Use this to recover the pure string. In other words,
        if a:mm then str(a):str.
        """
        return self.plain_str

    def __add__(self, other):
        # If other string is empty, don't let that take us out of math mode,
        # but do return a brand new MM.
        if len(other) == 0: return MM(self)
        # Sum of MM's is another MM.
        if isinstance(other, MM):
            return MM(self.plain_str + other.plain_str)
        # If added to plain string, surround with $'s and downgrade to plain.
        else:
            return f'${self}${other}'

    def __radd__(self, other):
        if len(other) == 0: return MM(self)
        if isinstance(other, MM):
            return MM(other.plain_str+self.plain_str)
        else:
            return f'{other}${self}$'

    def join(self, iterable):
        """
        If a:mm and L = [b0:mm, ..., bk:mm] then a.join(L):mm.
        """
        s = ''
        first = True
        for t in iterable:
            if first:
                s = t
                first = False
            else:
                s += self + t
        return s


def bracket_expression(style, expression):
    """
    Put brackets of a given style around a given expression.

    :param style: a bracket style; if invalid, expression will be returned
        unchanged.
    :param expression: the expression (str) to be bracketed

    :return: the modified expression
    """
    B = {
        'round': ("(", ")"),
        'square': (r'\[', r'\]'),
        'curly': (r'\lbrace', r'\rbrace'),
    }.get(style)
    if B is not None:
        expression = MM(r'\left') + MM(B[0]) + expression + MM(r'\right') + MM(B[1])
    return expression


def mbox(s):
    return s if isinstance(s, MM) else MM(r'\mbox{' + s + '}')


def wrap(obj):
    """
    Wrap an object, as needed, to make it a Formal object.
    """
    from stylgebra.expressions import FormalInteger, FormalQuotient, FormalEllipsis, FormalString
    if isinstance(obj, int):
        obj = FormalInteger(obj)
    elif isinstance(obj, SympyInteger):
        obj = FormalInteger(int(obj))
    elif isinstance(obj, SympyRational):
        n = int(obj.numerator)
        d = int(obj.denominator)
        obj = FormalQuotient(n, d)
    elif obj == Ellipsis:
        obj = FormalEllipsis()
    elif isinstance(obj, str):
        obj = FormalString(obj)
    return obj


class FormOpt:
    """
    Options for the 'form' property which is a part of the mode definition
    for almost all Formal classes.
    """
    NAME = 'name'
    VALUE = 'value'
    SYMBOLIC = 'symbolic'
    VERBAL = 'verbal'
    # Convenience for classes where 'value' and 'symbolic' are synonymous:
    valsym = [VALUE, SYMBOLIC]


EXPR_PATH_DELIMITER = '-'

# Check that a selector chain is at least a space-separated list
# of nonempty strings, starting with at least one nonempty string.
BASIC_SEL_CHAIN_PATTERN = re.compile(r'^\S+(\s+\S+)*$')


class ExpressionPath:
    """
    In an expression tree, Formal objects are the nodes, while "roles" are the edges.

    By a "role" we mean a string defining the role a subnode is playing
    relative to its parent node. For example, within a FormalPower there are
    two roles: "base" and "power". Within a FormalSum there is one role for
    each term, with the form `"term%d" % i`.

    This class represents a path of alternating nodes and edges leading down from
    the root of an expression tree to a single node within it.

    Because such a path alternates (node, role, node, role, ...), and because it
    always begins with a node and ends with a role, it always contains an equal
    number of nodes and roles.
    """

    def __init__(self, ancestors=None, roles=None, parent=None, role=None):
        self.ancestors = ancestors or []
        self.roles = roles or []
        if parent is not None and role is not None:
            self.append(parent, role)

    def append(self, parent, role):
        self.ancestors.append(parent)
        self.roles.append(role)

    def rolepath(self):
        return EXPR_PATH_DELIMITER.join(self.roles)

    def pop(self):
        a = self.ancestors.pop()
        r = self.roles.pop()
        return a, r

    def __str__(self):
        s = ''
        for a, r in zip(self.ancestors, self.roles):
            s += f'({a})-{r}-'
        return s

    def __bool__(self):
        return len(self) > 0

    def __len__(self):
        return len(self.roles)

    def __getitem__(self, key):
        # See <https://stackoverflow.com/a/16033058>
        if isinstance(key, slice):
            A = self.ancestors[key]
            R = self.roles[key]
            return ExpressionPath(ancestors=A, roles=R)
        return self.ancestors[key], self.roles[key]

    def __add__(self, other):
        return ExpressionPath(
            ancestors=self.ancestors + other.ancestors,
            roles=self.roles + other.roles
        )


class PathMatch:

    def __init__(self, end_pattern, exp_path):
        """
        :param end_pattern: string giving a desired final portion of a node path, with
            path segments separated by hyphens `-`.
            May use bracket notation `[i]` to indicate numeric matching groups.
        :param exp_path: the ExpressionPath leading up to the present point.

        For example, suppose

            end_pattern = 'cat-foo[i]bar[j]'
            rolepath = 'baz3-cat-foo17bar81'

        then...

        """
        p = '(?:^|-)%s$' % re.sub(r'\[([a-zA-Z]\w*)\]', r'(?P<\1>\\d+)', end_pattern)
        rolepath = exp_path.rolepath()
        m = re.search(p, rolepath)
        if m is None:
            self.matched = False
        else:
            self.matched = True
            self.indices = {k:int(v) for k, v in m.groupdict().items()}
            n = len(end_pattern.split('-'))
            self.num_segments_matched = n
            self.rolepath_prefix = rolepath[:m.span()[0]]
            self.exp_path_prefix = exp_path[:] if n == 1 else exp_path[:-n+1]


class Formal:

    def __init__(self, value=None, name=None, id=None, mode=None):
        r"""
        :param value: if provided, this should be an object that represents or defines
            the actual (numerical or other) value of the object in question. Not all
            objects will have a value.

        :param name: this should be a string giving the LaTeX you would use (without
            dollar signs) for the name by which you are calling this object within
            a given discussion.

        :param id: optional string with which to identify this object in formatting
            rules. If not provided, will be set equal to name.

        :param mode: this will be the initial default mode for governing the
            execution of the `format` method. It can be updated at any time using
            the `mode` method. If `None` is passed, we change this to an empty
            dictionary.
        """
        self._value = value
        self._name = name
        self._id = id or name
        self._mode = mode if mode is not None else {}

    def format(self, mode=None, rules=None, path=None, modifier=None):
        return ''

    def value(self, *args):
        if args:
            self._value = args[0]
        else:
            return self._value

    def name(self, *args):
        if args:
            self._name = args[0]
        else:
            return self._name

    def id(self, *args):
        if args:
            self._id = args[0]
        else:
            return self._id

    def mode(self, *args):
        if args:
            self._mode = args[0]
        else:
            return self._mode

    def decideMode(self, mode=None, rules=None, path=None, modifier=None):
        """
        Figure out which mode this object should be using during its `format` method.

        If a `mode` is given directly, take that.
        If not, try to find a mode in the rules.
        If that fails too, use this object's current default mode.
        However, in all these cases, if a modifier was given, apply it
            before returning. If the mode and modifier are both dictionaries,
            the mode will be upated by the modifier; otherwise the modifier
            should be a callable, and we apply it to the mode.
        """
        # If the mode is not given explicitly, we try to infer it from the rules.
        if mode is None and rules is not None:
            mode = self.searchRules(rules, path)
        # If we still don't have a mode, use our current default.
        if mode is None:
            mode = self.mode()
        # Modify?
        if modifier is not None:
            if isinstance(modifier, dict) and isinstance(mode, dict):
                mode = {**mode, **modifier}
            elif callable(modifier):
                mode = modifier(mode)
        return mode

    def searchRules(self, rules, path):
        """
        Search a (ordered) dictionary of rules for one that matches this object
        at the given expression path.

        Return the mode for the first rule that matches, or None.

        If a mode matches and is callable, we return the result of calling it
        at the indices matched by the rule.

        :param rules: should be a dictionary-like object. Keys are strings giving selectors,
            and values are the format modes that should be used if the rules match.

            The rules will be processed in order, and the first match will be accepted.
            So from Python 3.6 and onward, you can just use an ordinary dictionary, since these
            are now ordered. Before that, you should use collections.OrderedDict.
            See <https://stackoverflow.com/a/39980744>

        :param path: the path of the object; must be a string

        :return: the mode of the first rule that matched, or None
        """
        if path is None: path = ExpressionPath()
        for sel_chain, mode in rules.items():
            if not BASIC_SEL_CHAIN_PATTERN.match(sel_chain):
                raise FormalExcep('Malformed selector: %s' % sel_chain)
            selectors = sel_chain.split()
            indices = {}
            if self.matchRule(selectors, path, indices):
                if callable(mode):
                    arg_names = inspect.getfullargspec(mode).args
                    args = [indices.get(a) for a in arg_names]
                    mode = mode(*args)
                return mode
        return None

    def matchRule(self, selectors, path, indices, deferOnFail=False):
        """
        :param selectors: list of selectors remaining to be matched.
        :param path: ExpressionPath leading to this node.
        :param indices: a dictionary to be updated in-place with any indices
            matched in path segments.
        :param deferOnFail: this says what to do if we don't match: if True,
            ask parent node to make the same match; if False, just give up.

        :return: boolean saying whether we matched the rule or not.

        Selector formats:

            #id: match the id of the element

            type: match the "type" of the element, where "type" means the
                class name WITHOUT the leading "Formal".
                For example, use `Sum` to match any FormalSum object.

            @path: match the path of the element

                * path is absolute if selector comes first in the rule;
                  otherwise, path is relative to path of element matching
                  preceding selector.

                * Use bracket notation to match paths with indices.
                  E.g.
                        `term[i]-factor[j]`
                  will match
                        `term3-factor2`

                * When using bracket notation, use lambda abstraction when
                  defining the mode, using same variable names from the brackets.

                  Continuing the example above, could use

                    lambda i, j: {
                        'alpha': i + j
                    }

                  to substitute the sum of indices i and j for the value of a
                  formal variable of id 'alpha'.
        """
        sel = selectors.pop()
        defer = True
        next_path = None
        if sel[0] == "#":
            # id
            match = (self.id() == sel[1:])
        elif sel[0] == "@":
            # path
            pm = PathMatch(sel[1:], path)
            match = pm.matched
            if match:
                indices.update(pm.indices)
                next_path = pm.exp_path_prefix
                defer = False
        else:
            # type
            t = self.__class__.__name__
            if t.startswith('Formal'): t = t[6:]
            match = (t == sel)

        if not match:
            selectors.append(sel)

        if next_path is None:
            next_path = path[:]

        if (selectors and match) or (deferOnFail and not match):
            if next_path:
                ancestor, role = next_path.pop()
                return ancestor.matchRule(selectors, next_path, indices, deferOnFail=defer)
            else:
                return False
        else:
            return match

    def fwd(self, obj, role, rules, path, modifier=None, mode=None):
        """
        This is a convenience method with which to run a subobject's own
        format method and return the output.

        Given our expression path, plus the string naming the subobject's role,
        we assemble the subobject's expression path, and forward other relevant args.
        """
        if path is None: path = ExpressionPath()
        path += ExpressionPath(parent=self, role=role)
        return obj.format(mode=mode, rules=rules, path=path, modifier=modifier)

    def tryfwd(self, obj, role, rules, path, modifier=None, mode=None, mathmode=True):
        """
        Convenience method for forwarding when possible, else defaulting to
        mere string formatting.

        mathmode: Here you can say whether, in the case that the object is not
            a Formal subclass and must be converted to a string, it should be
            considered as being in math mode.

        All other parameters are as in the `fwd` method.
        """
        if isinstance(obj, Formal):
            return self.fwd(obj, role, rules, path, modifier=modifier, mode=mode)
        elif mathmode:
            return MM(obj)
        else:
            return str(obj)

    def getFwdMode(self, obj, role, rules, path, modifier=None, mode=None):
        """
        Return the mode that would be used by obj if we called `fwd` with the same args.
        """
        if path is None: path = ExpressionPath()
        path += ExpressionPath(parent=self, role=role)
        return obj.decideMode(mode=mode, rules=rules, path=path, modifier=modifier)

    ##########################################################################
    # Combining Operators

    def _(self, other):
        """
        Subscript
        """
        from stylgebra.expressions import FormalSubscripted
        other = wrap(other)
        return FormalSubscripted(self, other)

    def sub(self, other):
        """
        Alternative subscript, for symmetry with `sup`.
        """
        return self._(other)

    def __xor__(self, other):
        """
        Superscript
        """
        from stylgebra.expressions import FormalSuperscripted
        other = wrap(other)
        return FormalSuperscripted(self, other)

    def sup(self, other):
        """
        Alternative superscript
        """
        return self.__xor__(other)

    def __pow__(self, power, modulo=None):
        """
        Power
        """
        from stylgebra.expressions import FormalPower
        power = wrap(power)
        return FormalPower(self, power)

    def __neg__(self):
        from stylgebra.expressions import FormalSummand
        return -FormalSummand(self)

    def __add__(self, other):
        from stylgebra.expressions import FormalSummand
        other = wrap(other)
        return FormalSummand(self) + other

    def __sub__(self, other):
        from stylgebra.expressions import FormalSummand
        other = wrap(other)
        return FormalSummand(self) - other

    def __radd__(self, other):
        from stylgebra.expressions import FormalSummand
        other = wrap(other)
        return other + FormalSummand(self)

    def __rsub__(self, other):
        from stylgebra.expressions import FormalSummand
        other = wrap(other)
        return other - FormalSummand(self)

    def __mul__(self, other):
        from stylgebra.expressions import FormalProduct
        other = wrap(other)
        return FormalProduct([self, other])

    def __rmul__(self, other):
        other = wrap(other)
        from stylgebra.expressions import FormalProduct
        return FormalProduct([other, self])

    # Quotients for Python 2.x:
    def __div__(self, other):
        from stylgebra.expressions import FormalQuotient
        other = wrap(other)
        return FormalQuotient(self, other)

    def __rdiv__(self, other):
        from stylgebra.expressions import FormalQuotient
        other = wrap(other)
        return FormalQuotient(other, self)

    # Quotients for Python 3.x:
    def __truediv__(self, other):
        from stylgebra.expressions import FormalQuotient
        other = wrap(other)
        return FormalQuotient(self, other)

    def __rtruediv__(self, other):
        from stylgebra.expressions import FormalQuotient
        other = wrap(other)
        return FormalQuotient(other, self)


def withmode(format_method):
    """
    This function serves as a decorator for the `format` method of `Formal`
    subclasses. It serves to compute the mode, using the `decideMode` method,
    saving us from having to do that manually. In other words,

        @withmode
        def format(self, mode=None, rules=None, path=None, modifier=None):
            ...

    is the same as

        def format(self, mode=None, rules=None, path=None, modifier=None):
            mode = self.decideMode(mode, rules, path, modifier)
            ...

    It takes the same number of lines, but it is less tedious.
    """
    def f(formal_obj, mode=None, rules=None, path=None, modifier=None):
        mode = formal_obj.decideMode(mode, rules, path, modifier)
        return format_method(formal_obj, mode, rules, path, modifier)
    return f


class Infix:
    """
    This is a clever hack, thanks to
        <http://code.activestate.com/recipes/384122-infix-operators/>
    for turning any binary function into an infix operator.

    Given
        foo = lambda a, b: ...
    we can set
        infoo = Infix(foo)
    and then we will have
        a |infoo| b == foo(a, b)
    """
    
    def __init__(self, function):
        self.function = function
    
    def __ror__(self, other):
        return Infix(lambda x: self.function(other, x))
    
    def __or__(self, other):
        return self.function(other)
